import json
import tempfile
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import pytest
from git import Repo

TREE_SITTER_REPOS: dict[str, tuple[str, str, str]] = {
    "c_sharp": (
        "https://github.com/tree-sitter/tree-sitter-c-sharp.git",
        "362a8a41b265056592a0c3771664a21d23a71392",
        "tree-sitter-c-sharp",
    ),
    "dart": (
        "https://github.com/UserNobody14/tree-sitter-dart.git",
        "e81af6ab94a728ed99c30083be72d88e6d56cf9e",
        "tree-sitter-dart",
    ),
    "go": (
        "https://github.com/tree-sitter/tree-sitter-go.git",
        "3c3775faa968158a8b4ac190a7fda867fd5fb748",
        "tree-sitter-go",
    ),
    "hcl": (
        "https://github.com/tree-sitter-grammars/tree-sitter-hcl.git",
        "fad991865fee927dd1de5e172fb3f08ac674d914",
        "tree-sitter-hcl",
    ),
    "java": (
        "https://github.com/tree-sitter/tree-sitter-java.git",
        "94703d5a6bed02b98e438d7cad1136c01a60ba2c",
        "tree-sitter-java",
    ),
    "javascript": (
        "https://github.com/tree-sitter/tree-sitter-javascript.git",
        "3a837b6f3658ca3618f2022f8707e29739c91364",
        "tree-sitter-javascript",
    ),
    "json": (
        "https://github.com/tree-sitter/tree-sitter-json.git",
        "ee35a6ebefcef0c5c416c0d1ccec7370cfca5a24",
        "tree-sitter-json",
    ),
    "kotlin": (
        "https://github.com/tree-sitter-grammars/tree-sitter-kotlin.git",
        "77dd60ea0a9003ce062c9728a513ffe1aaff8c82",
        "tree-sitter-kotlin",
    ),
    "php": (
        "https://github.com/tree-sitter/tree-sitter-php.git",
        "43aad2b9a98aa8e603ea0cf5bb630728a5591ad8",
        "tree-sitter-php",
    ),
    "python": (
        "https://github.com/tree-sitter/tree-sitter-python.git",
        "bffb65a8cfe4e46290331dfef0dbf0ef3679de11",
        "tree-sitter-python",
    ),
    "ruby": (
        "https://github.com/tree-sitter/tree-sitter-ruby.git",
        "71bd32fb7607035768799732addba884a37a6210",
        "tree-sitter-ruby",
    ),
    "scala": (
        "https://github.com/tree-sitter/tree-sitter-scala.git",
        "d3b9553b47d7ab413a146c4b6498e527265b966e",
        "tree-sitter-scala",
    ),
    "swift": (
        "https://github.com/alex-pinkus/tree-sitter-swift.git",
        "0bea296efe6ea1f67338de26b56e389688d0acb8",
        "tree-sitter-swift",
    ),
    "tsx": (
        "https://github.com/tree-sitter/tree-sitter-typescript.git",
        "f975a621f4e7f532fe322e13c4f79495e0a7b2e7",
        "tree-sitter-typescript",
    ),
    "yaml": (
        "https://github.com/tree-sitter-grammars/tree-sitter-yaml",
        "3975596d84650e8d8ab8ef3128b1b7f0fec324f8",
        "tree-sitter-yaml",
    ),
}

STATIC_NODE_TYPES_DIR = Path(__file__).resolve().parents[2] / "blends" / "static" / "node_types"


def _get_fields(repo_url: str, commit_hash: str, repo_name: str) -> dict[str, list[str]]:
    with tempfile.TemporaryDirectory() as tmpfile:
        repo_path = Path(tmpfile) / repo_name
        repo = Repo.clone_from(repo_url, repo_path)
        repo.git.checkout(commit_hash)
        if repo_name == "tree-sitter-php":
            node_types = repo_path / "php" / "src" / "node-types.json"
        elif repo_name == "tree-sitter-typescript":
            node_types = repo_path / "tsx" / "src" / "node-types.json"
        else:
            node_types = repo_path / "src" / "node-types.json"
        with node_types.open(encoding="utf-8") as handle:
            fields: dict[str, list[str]] = {
                node["type"]: list(field_names)
                for node in json.load(handle)
                if (field_names := node.get("fields", {}).keys())
                if field_names
            }
            return fields


@pytest.mark.blends_test_group("tree_unittesting")
def test_verify_static_fields() -> None:
    def process_grammar(grammar: str, src: tuple[str, str, str]) -> None:
        path = STATIC_NODE_TYPES_DIR / f"{grammar}-fields.json"
        expected_fields = _get_fields(src[0], src[1], src[2])
        current_fields: dict[str, list[str]] = {}

        if path.is_file():
            with path.open(encoding="utf-8") as file:
                current_fields = json.load(file)

        if expected_fields != current_fields:
            with path.open(encoding="utf-8", mode="w") as file:
                json.dump(expected_fields, file, indent=2)

        assert expected_fields == current_fields

    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(process_grammar, grammar, src)
            for grammar, src in TREE_SITTER_REPOS.items()
        ]
        for future in futures:
            future.result()



from blends.models import (
    NId,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]

    constant_id = n_attrs["label_field_name"]
    expression = "::" + node_to_str(graph, constant_id)

    children: DirectChildren = {"expression_id": constant_id}
    if scope_id := n_attrs.get("label_field_scope"):
        children["object_id"] = scope_id

    return build_method_invocation_node(args, expression, children)



import pytest

from test.stack.test_helpers import (
    count_defs_refs,
    get_definition_precedences,
    resolve_symbol,
    setup_stack_resolution,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_linear_reassignment_returns_all_same_scope_defs() -> None:
    view, record = setup_stack_resolution("case_linear_reassignment.py")
    defs, refs = count_defs_refs(view, "x")

    assert defs == 2, f"Expected 2 defs of 'x', got {defs}"  # noqa: PLR2004
    assert refs == 1, f"Expected 1 ref to 'x', got {refs}"

    result = resolve_symbol(view, record, "x")
    assert len(result) == 2, (  # noqa: PLR2004
        f"Uniform precedence should return all same-scope defs, got {len(result)}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_parameter_then_reassignment_returns_all_defs() -> None:
    view, record = setup_stack_resolution("case_param_reassignment.py")
    defs, refs = count_defs_refs(view, "x")

    assert defs >= 2, f"Expected at least 2 defs of 'x', got {defs}"  # noqa: PLR2004
    assert refs >= 1, f"Expected at least 1 ref to 'x', got {refs}"

    result = resolve_symbol(view, record, "x")
    assert len(result) >= 2, (  # noqa: PLR2004
        f"Uniform precedence should return all same-scope defs, got {len(result)}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_different_symbols_not_affected() -> None:
    view, record = setup_stack_resolution("case_different_symbols.py")
    x_defs, x_refs = count_defs_refs(view, "x")
    y_defs, _y_refs = count_defs_refs(view, "y")

    assert x_defs == 1, f"Expected 1 def of 'x', got {x_defs}"
    assert x_refs == 1, f"Expected 1 ref to 'x', got {x_refs}"
    assert y_defs == 1, f"Expected 1 def of 'y', got {y_defs}"

    result = resolve_symbol(view, record, "x")
    assert len(result) == 1, f"'x' should resolve to 1 def, got {len(result)}"


@pytest.mark.blends_test_group("stack_unittesting")
def test_multiple_reassignments_returns_all_defs() -> None:
    view, record = setup_stack_resolution("case_multiple_reassignment.py")
    defs, refs = count_defs_refs(view, "x")

    assert defs == 3, f"Expected 3 defs of 'x', got {defs}"  # noqa: PLR2004
    assert refs == 1, f"Expected 1 ref to 'x', got {refs}"

    result = resolve_symbol(view, record, "x")
    assert len(result) == 3, (  # noqa: PLR2004
        f"Uniform precedence should return all same-scope defs, got {len(result)}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_uniform_precedence_for_parameter_and_assignments() -> None:
    view, _record = setup_stack_resolution("case_source_order_param.py")
    precedences = get_definition_precedences(view, "x")

    assert len(precedences) >= 2, (  # noqa: PLR2004
        f"Expected at least 2 defs of 'x', got {len(precedences)}"
    )
    assert all(p == 1 for p in precedences), (
        f"All same-scope binding precedences should be 1: {precedences}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_uniform_precedence_for_multiple_reassignments() -> None:
    view, _record = setup_stack_resolution("case_source_order_multi.py")
    precedences = get_definition_precedences(view, "x")

    assert len(precedences) >= 2, (  # noqa: PLR2004
        f"Expected at least 2 defs of 'x', got {len(precedences)}"
    )
    assert all(p == 1 for p in precedences), (
        f"All same-scope binding precedences should be 1: {precedences}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_uniform_precedence_for_different_symbols() -> None:
    view, _record = setup_stack_resolution("case_source_order_multi_symbols.py")

    x_precs = get_definition_precedences(view, "x")
    y_precs = get_definition_precedences(view, "y")

    assert len(x_precs) >= 2, f"Expected at least 2 defs of 'x', got {len(x_precs)}"  # noqa: PLR2004
    assert len(y_precs) >= 2, f"Expected at least 2 defs of 'y', got {len(y_precs)}"  # noqa: PLR2004
    assert all(p == 1 for p in x_precs), f"'x' precedences should all be 1: {x_precs}"
    assert all(p == 1 for p in y_precs), f"'y' precedences should all be 1: {y_precs}"


@pytest.mark.blends_test_group("stack_unittesting")
def test_nested_function_definition_returns_all_defs() -> None:
    view, record = setup_stack_resolution("case_source_order_nested_func.py")
    x_precs = get_definition_precedences(view, "x")

    assert len(x_precs) >= 2, f"Expected at least 2 defs of 'x', got {len(x_precs)}"  # noqa: PLR2004

    result = resolve_symbol(view, record, "x")
    assert len(result) >= 2, (  # noqa: PLR2004
        f"Uniform precedence should return all same-scope defs, got {len(result)}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_single_definition_no_reorder() -> None:
    view, record = setup_stack_resolution("case_different_symbols.py")

    x_precs = get_definition_precedences(view, "x")
    y_precs = get_definition_precedences(view, "y")

    assert len(x_precs) == 1, f"Expected 1 def of 'x', got {len(x_precs)}"
    assert len(y_precs) == 1, f"Expected 1 def of 'y', got {len(y_precs)}"

    result = resolve_symbol(view, record, "x")
    assert len(result) == 1, f"'x' should resolve to 1 def, got {len(result)}"


@pytest.mark.blends_test_group("stack_unittesting")
def test_conditional_branches_return_both_defs() -> None:
    view, record = setup_stack_resolution("case_conditional_branches.py")
    defs, refs = count_defs_refs(view, "x")

    assert defs == 2, f"Expected 2 defs of 'x', got {defs}"  # noqa: PLR2004
    assert refs == 1, f"Expected 1 ref to 'x', got {refs}"

    result = resolve_symbol(view, record, "x")
    assert len(result) == 2, (  # noqa: PLR2004
        f"Both branch defs should be returned, got {len(result)}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_parameter_shadows_module_level_variable() -> None:
    view, record = setup_stack_resolution("case_param_shadows_module.py")
    defs, refs = count_defs_refs(view, "x")

    assert defs >= 2, f"Expected at least 2 defs of 'x', got {defs}"  # noqa: PLR2004
    assert refs >= 1, f"Expected at least 1 ref to 'x', got {refs}"

    result = resolve_symbol(view, record, "x")
    assert len(result) == 1, f"Parameter should shadow module-level x, got {len(result)}"



from pathlib import Path

import pytest

from blends.models import Content, Graph, Language, SyntaxGraph
from blends.syntax.build_graph import generic, get_syntax_graph
from blends.syntax.models import SyntaxGraphArgs, SyntaxMetadata


@pytest.mark.blends_test_group("syntax_unittesting")
def test_get_syntax_graph_returns_graph_with_valid_input() -> None:
    content = Content(
        content_bytes=b"def hello():\n    return 'world'",
        content_str="def hello():\n    return 'world'",
        language=Language.PYTHON,
        path=Path("test.py"),
    )
    ast_graph = Graph()
    ast_graph.add_node("1", label_type="module")

    result = get_syntax_graph(ast_graph, content)

    assert result is not None
    assert isinstance(result, Graph)
    assert len(result.nodes) > 0


@pytest.mark.blends_test_group("syntax_unittesting")
def test_get_syntax_graph_with_metadata_flag() -> None:
    content = Content(
        content_bytes=b"x = 1",
        content_str="x = 1",
        language=Language.PYTHON,
        path=Path("test.py"),
    )
    ast_graph = Graph()
    ast_graph.add_node("1", label_type="module")

    result = get_syntax_graph(ast_graph, content, with_metadata=True)

    assert result is not None
    assert "0" in result.nodes
    assert result.nodes["0"]["label_type"] == "Metadata"
    assert result.nodes["0"]["path"] == "test.py"


@pytest.mark.blends_test_group("syntax_unittesting")
def test_get_syntax_graph_without_metadata_flag() -> None:
    content = Content(
        content_bytes=b"x = 1",
        content_str="x = 1",
        language=Language.PYTHON,
        path=Path("test.py"),
    )
    ast_graph = Graph()
    ast_graph.add_node("1", label_type="module")

    result = get_syntax_graph(ast_graph, content, with_metadata=False)

    assert result is not None
    assert "0" not in result.nodes or result.nodes.get("0", {}).get("label_type") != "Metadata"


@pytest.mark.blends_test_group("syntax_unittesting")
def test_generic_dispatches_to_reader() -> None:
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    ast_graph.add_node("1", label_type="module")

    metadata: SyntaxMetadata = {
        "class_path": [],
        "scope_stack": [],
    }

    args = SyntaxGraphArgs(
        generic=generic,
        path="test.py",
        language=Language.PYTHON,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id="1",
        metadata=metadata,
    )

    result = generic(args)

    assert isinstance(result, str)
    assert len(result) > 0


@pytest.mark.blends_test_group("syntax_unittesting")
def test_generic_handles_missing_dispatcher() -> None:
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    ast_graph.add_node("1", label_type="unknown_node_type_xyz")

    metadata: SyntaxMetadata = {
        "class_path": [],
        "scope_stack": [],
    }

    args = SyntaxGraphArgs(
        generic=generic,
        path="test.py",
        language=Language.PYTHON,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id="1",
        metadata=metadata,
    )

    result = generic(args)

    assert isinstance(result, str)
    assert len(result) > 0


@pytest.mark.blends_test_group("syntax_unittesting")
def test_get_syntax_graph_with_different_languages() -> None:
    languages_to_test = [
        Language.PYTHON,
        Language.JAVASCRIPT,
        Language.JAVA,
        Language.GO,
    ]

    for language in languages_to_test:
        content = Content(
            content_bytes=b"test",
            content_str="test",
            language=language,
            path=Path(f"test.{language.value}"),
        )
        ast_graph = Graph()
        ast_graph.add_node("1", label_type="module")

        result = get_syntax_graph(ast_graph, content)

        assert result is not None, f"Failed for language {language.value}"
        assert isinstance(result, Graph)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    value = ""
    text_id = match_ast_d(graph, args.n_id, "text")
    if text_id:
        value = node_to_str(graph, text_id)

    return build_string_literal_node(args, value)



from blends.models import (
    NId,
)
from blends.query import (
    get_node_by_path,
)
from blends.syntax.builders.import_global import (
    build_import_global_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    alias: str | None = None

    if (
        body_id := get_node_by_path(
            graph,
            args.n_id,
            "namespace_selectors",
            "arrow_renamed_identifier",
        )
    ) and (alias_n_id := nodes[body_id].get("label_field_alias")):
        alias = nodes[alias_n_id].get("label_text")
    full_import = node_to_str(graph, args.n_id)

    expression = (
        full_import.removeprefix("import").lstrip()
        if full_import.startswith("import")
        else full_import
    )

    return build_import_global_node(args, expression, set(), alias)



from blends.models import (
    NId,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    symbol = args.ast_graph.nodes[args.n_id].get("label_text")
    if not symbol:
        symbol = node_to_str(args.ast_graph, args.n_id)
    return build_symbol_lookup_node(args, symbol)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
    match_ast_d,
)
from blends.syntax.builders.pair import (
    build_pair_node,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    name_id = graph.nodes[args.n_id]["label_field_name"]
    type_id = None
    if type_annon_id := graph.nodes[args.n_id].get("label_field_type"):
        type_id = match_ast_d(graph, type_annon_id, "predefined_type")
        if not type_id:
            match_childs = match_ast(graph, type_annon_id, ":")
            type_id = match_childs.get("__0__")

    if type_id:
        return build_pair_node(args, name_id, type_id)
    return build_symbol_lookup_node(args, node_to_str(graph, name_id))



import time

import pytest

from blends.models import SyntaxGraph
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.forward_stitcher import ForwardStitcherConfig
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path import (
    PartialPath,
    PartialScopeStack,
    PartialSymbolStack,
)
from blends.stack.selection import PartialPathEdge
from blends.stack.single_file_resolver import (
    SingleFileResolutionRequest,
    SingleFileResolutionStats,
    resolve_definitions_single_file,
)
from blends.stack.view import StackGraphView
from test.stack.test_helpers import make_path, view_for_graph
from test.stack.test_helpers import record as make_record


def _path_with_edge(start: int, end: int, *, precedence: int) -> PartialPath:
    empty_symbols = PartialSymbolStack.empty()
    empty_scopes = PartialScopeStack.empty()
    return PartialPath(
        start_node_index=start,
        end_node_index=end,
        symbol_stack_precondition=empty_symbols,
        symbol_stack_postcondition=empty_symbols,
        scope_stack_precondition=empty_scopes,
        scope_stack_postcondition=empty_scopes,
        edges=(PartialPathEdge(source_node_index=start, precedence=precedence),),
    )


def _resolve_with_stats(
    view: StackGraphView, request: SingleFileResolutionRequest
) -> tuple[list[int], SingleFileResolutionStats]:
    result = resolve_definitions_single_file(view, request=request)
    assert isinstance(result, tuple)
    return result


def _build_view_with_defs() -> StackGraphView:
    graph = SyntaxGraph()
    graph.add_node("ref")
    graph.update_node_attribute("ref", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("ref", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("ref", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("def1")
    graph.update_node_attribute("def1", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("def1", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("def1", STACK_GRAPH_IS_DEFINITION, value=True)

    graph.add_node("def2")
    graph.update_node_attribute("def2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("def2", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("def2", STACK_GRAPH_IS_DEFINITION, value=True)

    return view_for_graph(graph)


@pytest.mark.blends_test_group("stack_unittesting")
def test_single_file_resolver_returns_empty_for_non_reference() -> None:
    graph = SyntaxGraph()
    graph.add_node("scope")
    graph.update_node_attribute("scope", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    view = view_for_graph(graph)
    scope_index = view.nid_to_index["scope"]

    request = SingleFileResolutionRequest(
        file_handle=1,
        record=make_record(()),
        ref_node_index=scope_index,
        include_stats=True,
    )
    result, stats = _resolve_with_stats(view, request)

    assert result == []
    assert stats.complete_paths_found == 0


@pytest.mark.blends_test_group("stack_unittesting")
def test_single_file_resolver_respects_work_budget() -> None:
    view = _build_view_with_defs()
    ref_index = view.nid_to_index["ref"]
    def_index = view.nid_to_index["def1"]

    record = make_record((make_path(ref_index, def_index),))
    request = SingleFileResolutionRequest(
        file_handle=1,
        record=record,
        ref_node_index=ref_index,
        include_stats=True,
        config=ForwardStitcherConfig(max_work_per_phase=0, max_phases=1),
    )

    result, stats = _resolve_with_stats(view, request)

    assert result == []
    assert stats.phases_executed == 1
    assert stats.candidates_considered == 0


@pytest.mark.blends_test_group("stack_unittesting")
def test_single_file_resolver_respects_deadline() -> None:
    view = _build_view_with_defs()
    ref_index = view.nid_to_index["ref"]
    def_index = view.nid_to_index["def1"]

    record = make_record((make_path(ref_index, def_index),))
    request = SingleFileResolutionRequest(
        file_handle=1,
        record=record,
        ref_node_index=ref_index,
        include_stats=True,
        config=ForwardStitcherConfig(deadline=time.monotonic() - 1),
    )

    result, stats = _resolve_with_stats(view, request)

    assert result == []
    assert stats.cancelled


@pytest.mark.blends_test_group("stack_unittesting")
def test_single_file_resolver_stitches_across_two_paths() -> None:
    graph = SyntaxGraph()
    graph.add_node("ref")
    graph.update_node_attribute("ref", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("ref", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("ref", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("join")
    graph.update_node_attribute("join", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)

    graph.add_node("def")
    graph.update_node_attribute("def", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("def", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("def", STACK_GRAPH_IS_DEFINITION, value=True)

    view = view_for_graph(graph)
    ref_index = view.nid_to_index["ref"]
    join_index = view.nid_to_index["join"]
    def_index = view.nid_to_index["def"]

    record = make_record((make_path(ref_index, join_index), make_path(join_index, def_index)))
    request = SingleFileResolutionRequest(
        file_handle=1,
        record=record,
        ref_node_index=ref_index,
    )

    result = resolve_definitions_single_file(view, request=request)

    assert result == [def_index]


@pytest.mark.blends_test_group("stack_unittesting")
def test_single_file_resolver_prunes_shadowed_candidates() -> None:
    view = _build_view_with_defs()
    ref_index = view.nid_to_index["ref"]
    def1_index = view.nid_to_index["def1"]
    def2_index = view.nid_to_index["def2"]

    record = make_record(
        (
            _path_with_edge(ref_index, def1_index, precedence=0),
            _path_with_edge(ref_index, def2_index, precedence=1),
        )
    )
    request = SingleFileResolutionRequest(
        file_handle=1,
        record=record,
        ref_node_index=ref_index,
    )

    result = resolve_definitions_single_file(view, request=request)

    assert result == [def2_index]



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.return_statement import (
    build_return_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    match = match_ast(args.ast_graph, args.n_id, "return", "yield", ";")
    min_match_count_for_selector_return = 5
    min_match_count_for_value_return = 3
    if (
        len(match) > min_match_count_for_selector_return
        and (n_id := match.get("__1__"))
        and args.ast_graph.nodes[n_id]["label_type"] == "selector"
    ):
        return build_return_node(args, value_id=n_id)
    if len(match) > min_match_count_for_value_return:
        return build_return_node(args, value_id=match["__0__"])
    return build_return_node(args, None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    pred_ast,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]

    pattern_id = n_attrs.get("label_field_pattern") or n_attrs.get("label_field_name")
    var_name = None
    if pattern_id:
        if args.ast_graph.nodes[pattern_id]["label_type"] == "object_pattern":
            var_name = node_to_str(args.ast_graph, pattern_id)[1:-1]
        else:
            var_name = node_to_str(args.ast_graph, pattern_id)

    type_id = n_attrs.get("label_field_type")
    var_type = None
    if type_id:
        var_type = node_to_str(args.ast_graph, type_id).replace(":", "")

    def_value = n_attrs.get("label_field_value")

    if not var_name or "," not in var_name:
        return build_parameter_node(
            args=args,
            variable=var_name,
            variable_type=var_type,
            value_id=def_value,
        )

    var_ids = (
        [
            _id
            for _id in adj_ast(args.ast_graph, pattern_id)
            if args.ast_graph.nodes[_id]["label_type"] not in {"{", "}", ","}
        ]
        if pattern_id
        else []
    )
    var_list = var_name.split(",") if var_name else []
    # As it's necessary to dynamically generate parameters in this case,
    # we have to save the last parameter as a return.
    # Otherwise, the reader will generate an extra (unnecessary) node.
    for var, var_id in zip(var_list[:-1], var_ids[:-1], strict=False):
        args.syntax_graph.add_edge(pred_ast(args.ast_graph, args.n_id)[0], var_id, label_ast="AST")
        build_parameter_node(
            args=args,
            variable=var,
            variable_type=var_type,
            value_id=def_value,
            variable_id=var_id,
        )
    return build_parameter_node(
        args=args,
        variable=var_list[-1] if len(var_list) > 1 else None,
        variable_type=var_type,
        value_id=def_value,
    )



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.models import (
    SyntaxCfgArgs,
)


def build(args: SyntaxCfgArgs) -> NId:
    al_id = args.graph.nodes[args.n_id].get("arguments_id")
    if al_id and match_ast_d(args.graph, al_id, "MethodDeclaration"):
        args.graph.add_edge(
            args.n_id,
            args.generic(args.fork(al_id, args.nxt_id)),
            label_cfg="CFG",
        )
    el_id = args.graph.nodes[args.n_id].get("expression_id")
    if (
        el_id
        and args.graph.nodes[el_id]["label_type"] == "MemberAccess"
        and (mi_id := match_ast_d(args.graph, el_id, "MethodInvocation", 1))
    ):
        args.graph.add_edge(
            args.n_id,
            args.generic(args.fork(mi_id, args.nxt_id)),
            label_cfg="CFG",
        )
    if bl_id := args.graph.nodes[args.n_id].get("block_id") or match_ast_d(
        args.graph, args.n_id, "ExecutionBlock"
    ):
        args.graph.add_edge(
            args.n_id,
            args.generic(args.fork(bl_id, args.nxt_id)),
            label_cfg="CFG",
        )
    if args.nxt_id:
        args.graph.add_edge(args.n_id, args.nxt_id, label_cfg="CFG")

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = args.ast_graph.nodes[args.n_id]

    ints = match_ast_group_d(graph, args.n_id, "template_substitution")
    int_subs = [
        child
        for ts_id in ints
        for child in adj_ast(graph, ts_id)
        if graph.nodes[child]["label_type"] not in {"$", "{", "}"}
    ]

    if int_subs:
        return build_string_literal_node(args, n_attrs["label_text"], iter(int_subs))
    return build_string_literal_node(args, n_attrs["label_text"])



from collections.abc import (
    Iterator,
)

from blends.models import (
    Language,
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_execution_block_node(args: SyntaxGraphArgs, c_ids: Iterator[NId]) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="ExecutionBlock",
    )
    for c_id in c_ids:
        if (
            args.language != Language.DART
            and args.ast_graph.nodes[c_id]["label_type"] == "expression_statement"
        ):
            child_id = adj_ast(args.ast_graph, c_id)[0]
            args.syntax_graph.add_edge(
                args.n_id,
                args.generic(args.fork_n_id(child_id)),
                label_ast="AST",
            )
        else:
            args.syntax_graph.add_edge(
                args.n_id,
                args.generic(args.fork_n_id(c_id)),
                label_ast="AST",
            )

    return args.n_id



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_if_node(
    args: SyntaxGraphArgs,
    condition_id: NId,
    true_id: NId | None = None,
    false_id: NId | None = None,
    initializer: NId | None = None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        condition_id=condition_id,
        label_type="If",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(condition_id)),
        label_ast="AST",
    )

    if true_id:
        args.syntax_graph.update_node_attribute(args.n_id, "true_id", true_id)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(true_id)),
            label_ast="AST",
        )

    if false_id:
        args.syntax_graph.update_node_attribute(args.n_id, "false_id", false_id)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(false_id)),
            label_ast="AST",
        )

    if initializer:
        args.syntax_graph.update_node_attribute(args.n_id, "initializer", initializer)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(initializer)),
            label_ast="AST",
        )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.spread_element import build_spread_element_node
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    match = match_ast(args.ast_graph, args.n_id, "**")
    identifier_id = match.get("__0__")
    if not identifier_id:
        identifier_id = adj_ast(args.ast_graph, args.n_id)[0]
    return build_spread_element_node(args, identifier_id)



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    n_attrs = args.graph.nodes[args.n_id]
    c_danger = generic(args.fork_n_id(n_attrs["constructor_id"])).danger
    if al_id := n_attrs.get("arguments_id"):
        al_danger = generic(args.fork_n_id(al_id)).danger
    else:
        al_danger = False

    args.evaluation[args.n_id] = c_danger or al_danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from collections.abc import (
    Callable,
)

from blends.models import (
    NId,
)
from blends.syntax.cfg.dispatchers import (
    await_expression_node,
    connect_to_block,
    connect_to_declarations,
    connect_to_next,
    end_node,
    if_node,
    method_invocation_node,
    multi_path,
    object_node,
    pair_node,
    step_by_step,
    switch_body,
    switch_section,
    variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxCfgArgs,
)

DISPATCHERS: dict[str, Callable[[SyntaxCfgArgs], NId]] = {
    "Class": connect_to_block.build,
    "DoStatement": connect_to_block.build,
    "ElseClause": connect_to_block.build,
    "ForEachStatement": connect_to_block.build,
    "ForStatement": connect_to_block.build,
    "MethodDeclaration": connect_to_block.build,
    "Namespace": connect_to_block.build,
    "SwitchStatement": connect_to_block.build,
    "UsingStatement": connect_to_block.build,
    "WhileStatement": connect_to_block.build,
    "ArgumentList": multi_path.build,
    "BinaryOperation": multi_path.build,
    "DeclarationBlock": multi_path.build,
    "File": multi_path.build,
    "TryStatement": multi_path.build,
    "ClassBody": multi_path.build,
    "SwitchBody": switch_body.build,
    "SwitchSection": switch_section.build,
    "ExecutionBlock": step_by_step.build,
    "ParameterList": step_by_step.build,
    "If": if_node.build,
    "TernaryOperation": if_node.build,
    "MethodInvocation": method_invocation_node.build,
    "Annotation": connect_to_next.build,
    "Argument": connect_to_next.build,
    "ArrayInitializer": connect_to_next.build,
    "Assignment": connect_to_next.build,
    "Attribute": connect_to_next.build,
    "AwaitExpression": await_expression_node.build,
    "Break": connect_to_next.build,
    "CatchClause": connect_to_next.build,
    "CatchDeclaration": connect_to_next.build,
    "Comment": connect_to_next.build,
    "Continue": connect_to_next.build,
    "Debugger": connect_to_next.build,
    "ElementAccess": connect_to_next.build,
    "ExpressionStatement": step_by_step.build,
    "FinallyClause": connect_to_next.build,
    "JsxElement": connect_to_next.build,
    "Literal": connect_to_next.build,
    "MemberAccess": connect_to_next.build,
    "MissingNode": connect_to_next.build,
    "Modifiers": connect_to_next.build,
    "NamedArgument": connect_to_next.build,
    "NewExpression": connect_to_next.build,
    "Object": object_node.build,
    "ObjectCreation": connect_to_next.build,
    "Pair": pair_node.build,
    "Parameter": connect_to_next.build,
    "ParenthesizedExpression": connect_to_next.build,
    "ReservedWord": connect_to_next.build,
    "RestPattern": connect_to_next.build,
    "Selector": connect_to_next.build,
    "SpreadElement": connect_to_next.build,
    "SymbolLookup": connect_to_next.build,
    "This": connect_to_next.build,
    "ThrowStatement": connect_to_next.build,
    "UnaryExpression": connect_to_next.build,
    "Yield": connect_to_next.build,
    "VariableDeclaration": variable_declaration_node.build,
    "Return": connect_to_declarations.build,
    "Export": connect_to_declarations.build,
    "Import": connect_to_next.build,
    "ModuleImport": end_node.build,
}



import pytest

from blends.stack import (
    MultiFileResolutionRequest,
    resolve_definitions_multi_file,
)
from blends.stack.forward_stitcher import ForwardStitcherConfig
from blends.stack.partial_path_database import PartialPathDatabase
from test.stack.test_helpers import (
    QUERY_MAX_PHASES,
    QUERY_MAX_WORK,
    find_symbol_nodes,
    setup_multi_file_database,
)

_MF_CORE = "mf_core.py"
_MF_TRANSPORT = "mf_transport.py"
_MF_PROCESSOR = "mf_processor.py"
_FH_CORE = 1
_FH_TRANSPORT = 2
_FH_PROCESSOR = 4

_CONFIG = ForwardStitcherConfig(
    max_phases=QUERY_MAX_PHASES,
    max_total_work=QUERY_MAX_WORK,
)


def _resolve(symbol: str, *, from_fh: int, views: dict, database: PartialPathDatabase) -> list:  # type: ignore[type-arg]
    view = views[from_fh]
    symbol_id = view.symbols.symbol_to_id.get(symbol)
    assert symbol_id is not None, f"Symbol '{symbol}' not found in view for fh={from_fh}"
    ref_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=False)
    assert ref_nodes, f"No reference nodes for '{symbol}' in fh={from_fh}"
    result = resolve_definitions_multi_file(
        view,
        database,
        request=MultiFileResolutionRequest(
            file_handle=from_fh,
            ref_node_index=ref_nodes[0],
            config=_CONFIG,
        ),
        views=views,
    )
    assert isinstance(result, list)
    return result


def _assert_resolves_to(
    symbol: str,
    *,
    from_fh: int,
    expected_fh: int,
    views: dict,  # type: ignore[type-arg]
    database: PartialPathDatabase,
) -> None:
    result = _resolve(symbol, from_fh=from_fh, views=views, database=database)
    file_handles = [fh for fh, _ in result]
    assert expected_fh in file_handles, (
        f"'{symbol}' from fh={from_fh} should resolve to fh={expected_fh}, got {file_handles}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_constant_imports_resolve_to_source() -> None:
    views, database = setup_multi_file_database({_MF_CORE: _FH_CORE, _MF_TRANSPORT: _FH_TRANSPORT})

    for symbol in ("SENTINEL", "MAX_POOL_SIZE", "HTTP_UNAVAILABLE"):
        _assert_resolves_to(
            symbol, from_fh=_FH_TRANSPORT, expected_fh=_FH_CORE, views=views, database=database
        )


@pytest.mark.blends_test_group("stack_unittesting")
def test_class_imports_resolve_to_source() -> None:
    views, database = setup_multi_file_database({_MF_CORE: _FH_CORE, _MF_TRANSPORT: _FH_TRANSPORT})

    for symbol in ("CoreError", "RetryPolicy", "ConnectionConfig"):
        _assert_resolves_to(
            symbol, from_fh=_FH_TRANSPORT, expected_fh=_FH_CORE, views=views, database=database
        )


@pytest.mark.blends_test_group("stack_unittesting")
def test_single_hop_to_defining_file() -> None:
    views, database = setup_multi_file_database(
        {_MF_TRANSPORT: _FH_TRANSPORT, _MF_PROCESSOR: _FH_PROCESSOR}
    )

    for symbol in ("TransportError", "ConnectionPool", "Request"):
        _assert_resolves_to(
            symbol, from_fh=_FH_PROCESSOR, expected_fh=_FH_TRANSPORT, views=views, database=database
        )


@pytest.mark.blends_test_group("stack_unittesting")
def test_two_hop_chain_resolves_to_origin() -> None:
    views, database = setup_multi_file_database(
        {_MF_CORE: _FH_CORE, _MF_TRANSPORT: _FH_TRANSPORT, _MF_PROCESSOR: _FH_PROCESSOR}
    )

    _assert_resolves_to(
        "SENTINEL", from_fh=_FH_PROCESSOR, expected_fh=_FH_CORE, views=views, database=database
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_direct_import_skips_intermediate() -> None:
    views, database = setup_multi_file_database(
        {_MF_CORE: _FH_CORE, _MF_TRANSPORT: _FH_TRANSPORT, _MF_PROCESSOR: _FH_PROCESSOR}
    )

    for symbol in ("RetryPolicy", "MAX_RETRIES"):
        _assert_resolves_to(
            symbol, from_fh=_FH_PROCESSOR, expected_fh=_FH_CORE, views=views, database=database
        )


_MF_BARE_MODULE = "mf_bare_module.py"
_FH_BARE_MODULE = 10

_MF_ORCHESTRATOR = "mf_orchestrator.py"
_FH_ORCHESTRATOR = 20

_MF_IMPORT_AS_CONSUMER = "mf_import_as_consumer.py"
_FH_IMPORT_AS_CONSUMER = 12


@pytest.mark.blends_test_group("stack_unittesting")
def test_aliased_import_resolves_to_source() -> None:
    views, database = setup_multi_file_database(
        {_MF_CORE: _FH_CORE, _MF_ORCHESTRATOR: _FH_ORCHESTRATOR}
    )
    _assert_resolves_to(
        "BUFFER_SIZE",
        from_fh=_FH_ORCHESTRATOR,
        expected_fh=_FH_CORE,
        views=views,
        database=database,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_bare_import_with_alias_is_import_only() -> None:
    views, database = setup_multi_file_database(
        {_MF_BARE_MODULE: _FH_BARE_MODULE, _MF_IMPORT_AS_CONSUMER: _FH_IMPORT_AS_CONSUMER}
    )
    result = _resolve("bm", from_fh=_FH_IMPORT_AS_CONSUMER, views=views, database=database)
    file_handles = [fh for fh, _ in result]
    assert _FH_IMPORT_AS_CONSUMER in file_handles, (
        f"import binding for 'bm' should appear in result, got {file_handles}"
    )
    assert _FH_BARE_MODULE not in file_handles, (
        f"mf_bare_module.py (fh={_FH_BARE_MODULE}) should not be reachable via "
        f"'import X as Z' without member-access resolution, got {file_handles}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_assignment_reexport_resolves_to_source() -> None:
    views, database = setup_multi_file_database(
        {_MF_PROCESSOR: _FH_PROCESSOR, _MF_ORCHESTRATOR: _FH_ORCHESTRATOR}
    )
    _assert_resolves_to(
        "OrchestratorError",
        from_fh=_FH_ORCHESTRATOR,
        expected_fh=_FH_PROCESSOR,
        views=views,
        database=database,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_absent_source_returns_import_binding() -> None:
    views, database = setup_multi_file_database({_MF_TRANSPORT: _FH_TRANSPORT})
    result = _resolve("SENTINEL", from_fh=_FH_TRANSPORT, views=views, database=database)

    file_handles = [fh for fh, _ in result]
    assert _FH_TRANSPORT in file_handles, (
        f"Import binding in mf_transport.py (fh={_FH_TRANSPORT}) should be returned as fallback, "
        f"got {file_handles}"
    )
    assert _FH_CORE not in file_handles, (
        f"mf_core.py (fh={_FH_CORE}) is not indexed, should not appear in result, "
        f"got {file_handles}"
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from blends.stack.policies.base import ExportPolicy
from blends.stack.scope_types import ScopeType

if TYPE_CHECKING:
    from blends.models import NId, SyntaxGraph


class JavaExportPolicy(ExportPolicy):
    def is_scope_exported(
        self,
        graph: SyntaxGraph,
        node_id: NId,
        scope_type: str | None,
    ) -> bool:
        if scope_type == ScopeType.FILE.value:
            return True
        if scope_type == ScopeType.CLASS.value:
            return self.get_visibility_modifier(graph, node_id) == "public"
        return False

    def is_definition_exported(
        self,
        graph: SyntaxGraph,
        node_id: NId,
        parent_scope_id: NId | None,
    ) -> bool:
        visibility = self.get_visibility_modifier(graph, node_id)
        if visibility != "public":
            return False
        if parent_scope_id is None:
            return True
        parent_is_exported = graph.nodes[parent_scope_id].get("stack_graph_is_exported", False)
        return bool(parent_is_exported)



import pytest

from blends.models import (
    Graph,
    SyntaxGraph,
)
from blends.stack.attributes import (
    STACK_GRAPH_IS_EXPORTED,
    STACK_GRAPH_KIND,
    STACK_GRAPH_PRECEDENCE,
    STACK_GRAPH_SCOPE,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_helpers import (
    drop_scopes_node_attributes,
    jump_to_node_attributes,
    pop_scoped_symbol_node_attributes,
    pop_symbol_node_attributes,
    push_scoped_symbol_node_attributes,
    push_symbol_node_attributes,
    root_node_attributes,
    scope_node_attributes,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_push_symbol_node() -> None:
    graph = Graph()
    graph.add_node("node_1", label_type="SymbolLookup")

    graph.update_node("node_1", push_symbol_node_attributes(symbol="foo", precedence=5))

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.PUSH_SYMBOL.value
    assert node_data[STACK_GRAPH_SYMBOL] == "foo"
    assert node_data[STACK_GRAPH_SCOPE] is None
    assert node_data[STACK_GRAPH_PRECEDENCE] == 5  # noqa: PLR2004
    assert node_data[STACK_GRAPH_IS_EXPORTED] is False
    assert node_data["label_type"] == "SymbolLookup"


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_push_symbol_node_default_precedence() -> None:
    graph = Graph()
    graph.add_node("node_1")

    graph.update_node("node_1", push_symbol_node_attributes(symbol="bar"))

    assert graph.nodes["node_1"][STACK_GRAPH_PRECEDENCE] == 0


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_push_scoped_symbol_node() -> None:
    graph = Graph()
    graph.add_node("node_1")
    graph.add_node("scope_1")

    graph.update_node(
        "node_1",
        push_scoped_symbol_node_attributes(symbol="foo", scope="scope_1", precedence=3),
    )

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value
    assert node_data[STACK_GRAPH_SYMBOL] == "foo"
    assert node_data[STACK_GRAPH_SCOPE] == "scope_1"
    assert node_data[STACK_GRAPH_PRECEDENCE] == 3  # noqa: PLR2004
    assert node_data[STACK_GRAPH_IS_EXPORTED] is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_pop_symbol_node() -> None:
    graph = Graph()
    graph.add_node("node_1", label_type="VariableDeclaration")

    graph.update_node("node_1", pop_symbol_node_attributes(symbol="x", precedence=10))

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.POP_SYMBOL.value
    assert node_data[STACK_GRAPH_SYMBOL] == "x"
    assert node_data[STACK_GRAPH_SCOPE] is None
    assert node_data[STACK_GRAPH_PRECEDENCE] == 10  # noqa: PLR2004
    assert node_data[STACK_GRAPH_IS_EXPORTED] is False
    assert node_data["label_type"] == "VariableDeclaration"


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_pop_scoped_symbol_node() -> None:
    graph = Graph()
    graph.add_node("node_1")

    graph.update_node("node_1", pop_scoped_symbol_node_attributes(symbol="y"))

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    assert node_data[STACK_GRAPH_SYMBOL] == "y"
    assert node_data[STACK_GRAPH_SCOPE] is None
    assert node_data[STACK_GRAPH_PRECEDENCE] == 0
    assert node_data[STACK_GRAPH_IS_EXPORTED] is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_scope_node_exported() -> None:
    graph = Graph()
    graph.add_node("node_1", label_type="File")

    graph.update_node("node_1", scope_node_attributes(is_exported=True))

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.SCOPE.value
    assert node_data[STACK_GRAPH_SYMBOL] is None
    assert node_data[STACK_GRAPH_SCOPE] is None
    assert node_data[STACK_GRAPH_PRECEDENCE] == 0
    assert node_data[STACK_GRAPH_IS_EXPORTED] is True
    assert node_data["label_type"] == "File"


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_scope_node_not_exported() -> None:
    graph = Graph()
    graph.add_node("node_1", label_type="Class")

    graph.update_node("node_1", scope_node_attributes(is_exported=False))

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.SCOPE.value
    assert node_data[STACK_GRAPH_IS_EXPORTED] is False
    assert node_data["label_type"] == "Class"


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_root_node() -> None:
    graph = Graph()
    graph.add_node("node_1")

    graph.update_node("node_1", root_node_attributes())

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.ROOT.value
    assert node_data[STACK_GRAPH_SYMBOL] is None
    assert node_data[STACK_GRAPH_SCOPE] is None
    assert node_data[STACK_GRAPH_PRECEDENCE] == 0
    assert node_data[STACK_GRAPH_IS_EXPORTED] is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_jump_to_node() -> None:
    graph = Graph()
    graph.add_node("node_1")

    graph.update_node("node_1", jump_to_node_attributes())

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.JUMP_TO.value
    assert node_data[STACK_GRAPH_SYMBOL] is None
    assert node_data[STACK_GRAPH_SCOPE] is None
    assert node_data[STACK_GRAPH_PRECEDENCE] == 0
    assert node_data[STACK_GRAPH_IS_EXPORTED] is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_mark_as_drop_scopes_node() -> None:
    graph = Graph()
    graph.add_node("node_1")

    graph.update_node("node_1", drop_scopes_node_attributes())

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.DROP_SCOPES.value
    assert node_data[STACK_GRAPH_SYMBOL] is None
    assert node_data[STACK_GRAPH_SCOPE] is None
    assert node_data[STACK_GRAPH_PRECEDENCE] == 0
    assert node_data[STACK_GRAPH_IS_EXPORTED] is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_helper_preserves_existing_attributes() -> None:
    graph = Graph()
    graph.add_node(
        "node_1",
        label_type="SymbolLookup",
        symbol="x",
        custom_attr="preserved",
    )

    graph.update_node("node_1", push_symbol_node_attributes(symbol="y"))

    node_data = graph.nodes["node_1"]
    assert node_data["label_type"] == "SymbolLookup"
    assert node_data["custom_attr"] == "preserved"
    assert node_data[STACK_GRAPH_SYMBOL] == "y"


@pytest.mark.blends_test_group("stack_unittesting")
def test_helper_overwrites_existing_stack_graph_attributes() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, "OldKind")
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 99)

    graph.update_node("node_1", push_symbol_node_attributes(symbol="foo", precedence=5))

    node_data = graph.nodes["node_1"]
    assert node_data[STACK_GRAPH_KIND] == StackGraphNodeKind.PUSH_SYMBOL.value
    assert node_data[STACK_GRAPH_PRECEDENCE] == 5  # noqa: PLR2004



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    childs = adj_ast(
        args.ast_graph,
        args.n_id,
    )
    invalid_types = {
        "assert",
        ",",
        ")",
        "(",
    }

    valid_childs = [
        _id for _id in childs if args.ast_graph.nodes[_id]["label_type"] not in invalid_types
    ]

    return build_expression_statement_node(args, iter(valid_childs))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    filtered_ids: list[NId] = []

    for _id in c_ids:
        n_attrs = graph.nodes[_id]
        if n_attrs.get("label_type") in {
            "import_statement",
            "import_from_statement",
        }:
            filtered_ids.extend(match_ast_group_d(args.ast_graph, _id, "dotted_name", depth=-1))
            if match_ast_d(graph, _id, "wildcard_import"):
                continue
            if module_n_id := n_attrs.get("label_field_module_name"):
                if module_n_id in filtered_ids:
                    filtered_ids.remove(module_n_id)
                elif d_name_n_id := match_ast_d(graph, module_n_id, "dotted_name"):
                    filtered_ids.remove(d_name_n_id)

            continue
        filtered_ids.append(_id)

    return build_file_node(args, iter(filtered_ids))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.member_access import (
    build_member_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)

    relative_n_id = c_ids[0]
    member_n_id = c_ids[-1]

    expression = graph.nodes[member_n_id].get("label_text", "")

    obj_id = match_ast_d(graph, relative_n_id, "self") or member_n_id

    return build_member_access_node(args, "self", expression, obj_id)



import pytest

TEST_GROUPS = {
    "all",
    "ast_unittesting",
    "content_unittesting",
    "query_unittesting",
    "stack_unittesting",
    "symbolic_evaluation_unittesting",
    "syntax_unittesting",
    "tree_unittesting",
    "functional",
}


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--blends-test-group",
        action="store",
        metavar="BLENDS_TEST_GROUP",
    )


def pytest_runtest_setup(item: pytest.Module) -> None:
    blends_test_group = item.config.getoption("--blends-test-group")

    if not blends_test_group:
        try:
            marker = next(x for x in item.iter_markers(name="blends_test_group"))
        except StopIteration as exc:
            exc_log = "blends-test-group not specified"
            raise ValueError(exc_log) from exc
        blends_test_group = marker.args[0]

    if blends_test_group != "all" and blends_test_group not in TEST_GROUPS:
        exc_log = f"blends-test-group must be one of: {TEST_GROUPS}, or all"
        raise ValueError(exc_log)

    runnable_groups = {mark.args[0] for mark in item.iter_markers(name="blends_test_group")}

    if not runnable_groups or runnable_groups - TEST_GROUPS:
        exc_log = f"blends-test-group must be one of: {TEST_GROUPS}"
        raise ValueError(exc_log)

    if blends_test_group != "all" and blends_test_group not in runnable_groups:
        pytest.skip(f"Requires blends test group in: {runnable_groups}")



from blends.models import (
    NId,
)
from blends.syntax.builders.binary_operation import build_binary_operation_node
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    left_id = args.ast_graph.nodes[args.n_id]["label_field_expr"]
    right_id = args.ast_graph.nodes[args.n_id]["label_field_type"]
    operator = "as_expression"
    return build_binary_operation_node(args, operator, left_id, right_id)



from collections.abc import (
    Iterator,
)

from blends.models import (
    Graph,
    NId,
)
from blends.query import (
    adj_ast,
)


def lazy_childs_text(graph: Graph, n_id: NId) -> Iterator[str]:
    for c_id in adj_ast(graph, n_id):
        yield from lazy_childs_text(graph, c_id)
    if "label_text" in graph.nodes[n_id]:
        yield graph.nodes[n_id]["label_text"]


def node_to_str(graph: Graph, n_id: NId, sep: str = "") -> str:
    return sep.join(lazy_childs_text(graph, n_id))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.try_statement import (
    build_try_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    block_id = graph.nodes[args.n_id].get("label_field_expr")
    if not block_id:
        block_id = adj_ast(graph, args.n_id)[0]
    return build_try_statement_node(args, block_id, None, None, None)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.object_creation import (
    build_object_creation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    node_attr = args.ast_graph.nodes[args.n_id]
    type_id = node_attr["label_field_type"]
    name = node_to_str(args.ast_graph, type_id)

    arguments_id = node_attr["label_field_arguments"]
    if "__0__" not in match_ast(args.ast_graph, arguments_id, "(", ")"):
        arguments_id = None
    init_node: NId | None = match_ast(args.ast_graph, args.n_id, "class_body").get("class_body")

    return build_object_creation_node(args, name, arguments_id, init_node)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    pred_ast,
)
from blends.syntax.builders.argument_list import (
    build_argument_list_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    invalid_types = {
        "echo",
        ",",
        ";",
    }

    nested_types = {"sequence_expression"}

    p_id = pred_ast(graph, args.n_id)[0]

    valid_childs: set[NId] = set()
    for c_id in adj_ast(args.ast_graph, p_id):
        if (label_type := graph.nodes[c_id].get("label_type")) in invalid_types:
            continue
        if label_type in nested_types:
            clean_nested_childs = [
                n_id
                for n_id in adj_ast(graph, c_id)
                if graph.nodes[n_id].get("label_type") not in invalid_types
            ]
            valid_childs.update(clean_nested_childs)
            continue
        valid_childs.add(c_id)

    return build_argument_list_node(args, iter(valid_childs))



from blends.models import (
    NId,
)
from blends.query import (
    get_nodes_by_path,
    match_ast_d,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    ListChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    body_id = (
        match_ast_d(args.ast_graph, args.n_id, "block")
        or match_ast_d(args.ast_graph, args.n_id, "infix_expression")
        or match_ast_d(args.ast_graph, args.n_id, "indented_block")
    )

    direct: DirectChildren = {}
    list_children: ListChildren = {}
    if body_id:
        direct["block_id"] = body_id
    if param_id := n_attrs.get("label_field_parameters"):
        direct["parameters_id"] = param_id
    elif condition_nodes := list(get_nodes_by_path(graph, args.n_id, "bindings", "binding")):
        list_children["parameters_ids"] = condition_nodes

    return build_method_declaration_node(args, "LambdaExpression", direct, list_children)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expr_id = graph.nodes[args.n_id]["label_field_name"]

    arguments_id = graph.nodes[args.n_id]["label_field_arguments"]
    if "__0__" not in match_ast(args.ast_graph, arguments_id, "(", ")"):
        arguments_id = None

    expr = node_to_str(graph, expr_id)
    children: DirectChildren = {"expression_id": expr_id}
    if arguments_id:
        children["arguments_id"] = arguments_id

    if object_id := graph.nodes[args.n_id].get("label_field_object"):
        children["object_id"] = object_id
        object_name = None
        if graph.nodes[object_id]["label_type"] == "object_creation_expression":
            object_name = node_to_str(graph, graph.nodes[object_id]["label_field_type"])
        return build_method_invocation_node(args, expr, children, object_name)

    return build_method_invocation_node(args, expr, children)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.throw import (
    build_throw_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    childs = match_ast(args.ast_graph, args.n_id, "throw", ";")
    expr_id = childs.get("__0__")
    return build_throw_node(args, expr_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.argument import (
    build_argument_node,
)
from blends.syntax.builders.named_argument import (
    build_named_argument_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    first_child, *other_childs = adj_ast(graph, args.n_id)

    if not other_childs:
        return args.generic(args.fork_n_id(first_child))

    if other_childs and (label_id := match_ast_d(graph, args.n_id, "label")):
        var_id = adj_ast(graph, label_id)[0]
        arg_name = node_to_str(graph, var_id)
        return build_named_argument_node(args, arg_name, other_childs[0])

    valid_childs = list(adj_ast(args.ast_graph, args.n_id))
    return build_argument_node(args, iter(valid_childs))



from itertools import (
    dropwhile,
)

from blends.ctx import ctx
from blends.models import (
    LANGUAGE_TO_EXTENSIONS,
    NId,
)
from blends.syntax.models import (
    FileInstanceData,
    FileStructData,
    SyntaxGraphArgs,
)
from blends.utilities.strings import (
    split_on_last_dot as split_last,
)


def add_class_to_metadata(args: SyntaxGraphArgs, name: str) -> None:
    parent_class = args.syntax_graph.nodes[NId(0)]["structure"]

    if args.metadata["class_path"]:
        for elem in args.metadata["class_path"]:
            parent_class = parent_class[elem]["data"]

        parent_class[name] = FileStructData(node=args.n_id, type="class", data={})

        args.metadata["class_path"].append(name)
    else:
        parent_class[name] = FileStructData(node=args.n_id, type="class", data={})
        args.metadata["class_path"] = [name]


def add_method_to_metadata(args: SyntaxGraphArgs, name: str) -> None:
    parent_class = args.syntax_graph.nodes[NId(0)]["structure"]

    for elem in args.metadata["class_path"]:
        parent_class = parent_class[elem]["data"]

    parent_class[name] = FileStructData(node=args.n_id, type="method", data=name)


def add_node_range_to_method(args: SyntaxGraphArgs, name: str) -> None:
    parent_class = args.syntax_graph.nodes[NId(0)]["structure"]
    for elem in args.metadata["class_path"]:
        parent_class = parent_class[elem]["data"]
    if method_dict := parent_class.get(name):
        method_dict["node_range"] = list(
            dropwhile(lambda x: x != args.n_id, list(args.syntax_graph.nodes)),
        )


def del_metadata_instance(args: SyntaxGraphArgs, var_id: NId, val_id: NId) -> None:
    val_attrs = args.syntax_graph.nodes[val_id]
    var_attrs = args.syntax_graph.nodes[var_id]

    var = var_attrs.get("symbol")

    current_class = args.metadata["class_path"][-1] if args.metadata["class_path"] else None
    class_instances = (
        args.syntax_graph.nodes[NId(0)]["instances"][current_class]
        if current_class and args.syntax_graph.nodes[NId(0)]["instances"]
        else None
    )

    if not (class_instances and var and class_instances.get(var)):
        return

    if not (
        val_attrs["label_type"] == "ObjectCreation"
        and val_attrs["name"] == class_instances[var]["object"]
    ):
        del class_instances[var]


def get_file_from_path(file_name: str) -> str | None:
    min_path_depth_for_file_search = 4
    if file_name.count("/") >= min_path_depth_for_file_search:
        for config_path in ctx.iterate_multi_paths():
            file_path = str(config_path)
            if file_name in file_path:
                return file_path
    return None


def add_instance_to_metadata(args: SyntaxGraphArgs, var_type: str, var_name: str) -> None:
    current_class = args.metadata["class_path"][-1]
    if current_class not in args.syntax_graph.nodes[NId(0)]["instances"]:
        args.syntax_graph.nodes[NId(0)]["instances"][current_class] = {}

    possible_path = var_type.replace(".", "/") + LANGUAGE_TO_EXTENSIONS[args.language][0]

    if path := get_file_from_path(possible_path):
        args.syntax_graph.nodes[NId(0)]["instances"][current_class][var_name] = FileInstanceData(
            object=split_last(var_type)[1],
            source=path,
            source_type="file_path",
        )

    for imported_package in args.syntax_graph.nodes[NId(0)]["imports"]:
        split_import = split_last(imported_package)
        split_var = split_last(var_type)
        match_names = {imported_package}
        if split_import[1] == "*":
            if split_import[0] == split_var[0]:
                match_names.add(var_type)
        else:
            match_names.add(split_import[1])
        if var_type in match_names:
            args.syntax_graph.nodes[NId(0)]["instances"][current_class][var_name] = (
                FileInstanceData(
                    object=var_type,
                    source=split_import[0],
                    source_type="package",
                )
            )



from blends.symbolic_evaluation.context.member import (
    solve as solve_member_access,
)
from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    expr_id = solve_member_access(args) or args.graph.nodes[args.n_id]["expression_id"]
    args.evaluation[args.n_id] = generic(args.fork_n_id(expr_id)).danger

    if args.method_evaluators and (method_evaluator := args.method_evaluators.get("member_access")):
        args.evaluation[args.n_id] = method_evaluator(args).danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.do_statement import (
    build_do_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.syntax.readers.constants import (
    C_SHARP_EXPRESSION,
    C_SHARP_STATEMENT,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = adj_ast(args.ast_graph, args.n_id)
    body_id = [_id for _id in c_ids if graph.nodes[_id]["label_type"] in C_SHARP_STATEMENT].pop()
    if graph.nodes[body_id]["label_type"] == "expression_statement":
        body_id = adj_ast(graph, body_id)[0]

    condition_node = [
        _id for _id in c_ids if graph.nodes[_id]["label_type"] in C_SHARP_EXPRESSION
    ].pop()

    return build_do_statement_node(args, body_id, condition_node)



import tempfile
from pathlib import Path

import pytest

from blends.content.content import (
    FileTooLargeError,
    _get_file_raw_content,
    get_content_by_path,
    get_language_by_path,
)
from blends.content.custom_parsers import custom_parsers_by_language
from blends.models import Content, Language


@pytest.mark.blends_test_group("content_unittesting")
def test_get_language_by_path() -> None:  # noqa: PLR0915
    assert get_language_by_path(Path("mono/csharp/Service.cs")) == Language.CSHARP
    assert get_language_by_path(Path("app/csharp/logic/Middleware.CS")) == Language.CSHARP
    assert get_language_by_path(Path("lib/dartWidget/src/Widget.dart")) == Language.DART
    assert get_language_by_path(Path("mobile/dart/App.DART")) == Language.DART
    assert get_language_by_path(Path("services/user/Api.java")) == Language.JAVA
    assert get_language_by_path(Path("integration/test/TestSuite.JAVA")) == Language.JAVA
    assert get_language_by_path(Path("tools/kotlin/MyApp.kt")) == Language.KOTLIN
    assert get_language_by_path(Path("client/ui/view/Widget.jsx")) == Language.JAVASCRIPT
    assert get_language_by_path(Path("tools/scripts/Utility.JS")) == Language.JAVASCRIPT
    assert get_language_by_path(Path("frontend/components/Button.ts")) == Language.TYPESCRIPT
    assert get_language_by_path(Path("src/project/module/file.py")) == Language.PYTHON
    assert get_language_by_path(Path("some/repo/logic/BackEnd.PY")) == Language.PYTHON
    assert get_language_by_path(Path("web/assets/scripts/main.js")) == Language.JAVASCRIPT
    assert get_language_by_path(Path("cmd/tools/GoTool.GO")) == Language.GO
    assert get_language_by_path(Path("backend/go/server.go")) == Language.GO
    assert get_language_by_path(Path("cloud/hcl/main.hcl")) == Language.HCL
    assert get_language_by_path(Path("terraform/modules/config.tf")) == Language.HCL
    assert get_language_by_path(Path("database/seed/data.json")) == Language.JSON
    assert get_language_by_path(Path("resources/config/Info.JSON")) == Language.JSON
    assert get_language_by_path(Path("api/php/entry.php")) == Language.PHP
    assert get_language_by_path(Path("src/domain/Example.KT")) == Language.KOTLIN
    assert get_language_by_path(Path("legacy/src/Index.PHP")) == Language.PHP
    assert get_language_by_path(Path("web/ruby/on_rails/model.rb")) == Language.RUBY
    assert get_language_by_path(Path("scripts/util/Helper.RB")) == Language.RUBY
    assert get_language_by_path(Path("scala_project/src/main.scala")) == Language.SCALA
    assert get_language_by_path(Path("scalajs/app/Core.SCALA")) == Language.SCALA
    assert get_language_by_path(Path("ios/swift/ViewController.swift")) == Language.SWIFT
    assert get_language_by_path(Path("lib/module/Editor.tsx")) == Language.TYPESCRIPT
    assert get_language_by_path(Path("compiler/transformers/Adapter.TS")) == Language.TYPESCRIPT
    assert get_language_by_path(Path("macos/SwiftTypes/Type.SWIFT")) == Language.SWIFT
    assert get_language_by_path(Path("config/yaml/settings.yaml")) == Language.YAML
    assert get_language_by_path(Path("docker/compose/File.YML")) == Language.YAML
    assert get_language_by_path(Path("src/unknown/randomfile.unknown")) == Language.NOT_SUPPORTED
    assert get_language_by_path(Path("folder/with/no_extension")) == Language.NOT_SUPPORTED


@pytest.mark.blends_test_group("content_unittesting")
def test_get_file_raw_content_success() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.txt"
        test_content = b"Hello, World!"
        test_file.write_bytes(test_content)

        result = _get_file_raw_content(test_file)

        assert result == test_content


@pytest.mark.blends_test_group("content_unittesting")
def test_get_file_raw_content_with_size_limit() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.txt"
        test_content = b"Hello, World!"
        test_file.write_bytes(test_content)

        result = _get_file_raw_content(test_file, size=100)

        assert result == test_content


@pytest.mark.blends_test_group("content_unittesting")
def test_get_file_raw_content_file_too_large() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.txt"
        test_content = b"x" * 100
        test_file.write_bytes(test_content)

        with pytest.raises(FileTooLargeError):
            _get_file_raw_content(test_file, size=50)


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_success_python_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "module.py"
        test_content = "def hello():\n    print('world')"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.content_str == test_content
        assert result.content_bytes == test_content.encode("utf-8")
        assert result.language == Language.PYTHON
        assert result.path == test_file


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_success_javascript_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "script.js"
        test_content = "function test() { return true; }"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.content_str == test_content
        assert result.language == Language.JAVASCRIPT
        assert result.path == test_file


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_file_too_large() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "large.txt"
        large_content = "x" * 1000
        test_file.write_text(large_content, encoding="utf-8")

        result = get_content_by_path(test_file, max_size=100)

        assert result is None


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_file_not_found() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        non_existent_file = Path(temp_dir) / "nonexistent.py"

        result = get_content_by_path(non_existent_file)

        assert result is None


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_directory_instead_of_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        directory_path = Path(temp_dir)

        result = get_content_by_path(directory_path)

        assert result is None


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_custom_max_size() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.py"
        test_content = "print('test')"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file, max_size=50)

        assert result is not None
        assert result.content_str == test_content


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_not_supported_language() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "file.unknown"
        test_content = "some content"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.language == Language.NOT_SUPPORTED
        assert result.content_str == test_content


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_json_without_helm_template() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "config.json"
        test_content = '{"key": "value"}'
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.language == Language.JSON
        assert result.content_str == test_content


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_yaml_without_helm_template() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "config.yaml"
        test_content = "key: value\nother: data"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.language == Language.YAML
        assert result.content_str == test_content


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_json_helm_template_in_templates_directory() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        templates_dir = Path(temp_dir) / "templates"
        templates_dir.mkdir()
        test_file = templates_dir / "deployment.json"
        test_content = (
            '{"apiVersion": "v1", "kind": "Deployment", "metadata": {"name": "test"}, '
            '"spec": {"replicas": {{ .Values.replicas }}}}'
        )
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.language == Language.JSON


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_yaml_helm_template_in_templates_directory() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        templates_dir = Path(temp_dir) / "templates"
        templates_dir.mkdir()
        test_file = templates_dir / "deployment.yaml"
        test_content = (
            "apiVersion: v1\nkind: Deployment\nmetadata:\n  name: {{ .Values.name }}\n"
            "spec:\n  replicas: {{ .Values.replicas }}"
        )
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.language == Language.YAML


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_json_helm_template_not_in_templates_directory() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "deployment.json"
        test_content = (
            '{"apiVersion": "v1", "kind": "Deployment", "metadata": {"name": "test"}, '
            '"spec": {"replicas": {{ .Values.replicas }}}}'
        )
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.language == Language.JSON
        assert result.content_str == test_content


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_empty_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "empty.py"
        test_file.touch()

        result = get_content_by_path(test_file)

        assert result is not None
        assert isinstance(result, Content)
        assert result.content_str == ""
        assert result.content_bytes == b""
        assert result.language == Language.PYTHON


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_multiline_content() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "multiline.py"
        test_content = "line1\nline2\nline3\n"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert result.content_str == test_content
        assert result.content_bytes == test_content.encode("utf-8")


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_special_characters() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "special.py"
        test_content = "print('ñáéíóú中文🚀')"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file)

        assert result is not None
        assert result.content_str == test_content
        assert result.content_bytes == test_content.encode("utf-8")


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_exact_max_size() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "exact.py"
        content_len = 100
        test_content = "x" * content_len
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file, max_size=100)

        assert result is not None
        assert len(result.content_bytes) == content_len


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_one_byte_over_max_size() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "over.py"
        test_content = "x" * 101
        test_file.write_text(test_content, encoding="utf-8")

        result = get_content_by_path(test_file, max_size=100)

        assert result is None


@pytest.mark.blends_test_group("content_unittesting")
def test_get_content_by_path_all_supported_languages() -> None:
    language_extensions = {
        Language.PYTHON: ".py",
        Language.JAVASCRIPT: ".js",
        Language.TYPESCRIPT: ".ts",
        Language.JAVA: ".java",
        Language.KOTLIN: ".kt",
        Language.CSHARP: ".cs",
        Language.DART: ".dart",
        Language.GO: ".go",
        Language.HCL: ".hcl",
        Language.JSON: ".json",
        Language.PHP: ".php",
        Language.RUBY: ".rb",
        Language.SCALA: ".scala",
        Language.SWIFT: ".swift",
        Language.YAML: ".yaml",
        Language.NOT_SUPPORTED: ".unknown",
    }

    with tempfile.TemporaryDirectory() as temp_dir:
        for language, extension in language_extensions.items():
            test_file = Path(temp_dir) / f"test{extension}"
            test_content = f"content for {language.value}"
            test_file.write_text(test_content, encoding="utf-8")

            result = get_content_by_path(test_file)

            assert result is not None
            assert result.language == language
            assert result.content_str == test_content


@pytest.mark.blends_test_group("content_unittesting")
def test_helm_parser_validates_and_parses_all_templates() -> None:
    templates_dir = Path(__file__).parent.parent / "data" / "helm_parser" / "templates"

    assert templates_dir.exists(), f"Templates directory not found: {templates_dir}"

    template_files = list[Path](templates_dir.iterdir())
    assert template_files, f"No template files found in {templates_dir}"

    for template_path in template_files:
        if not template_path.is_file():
            continue

        language = get_language_by_path(template_path)
        content_bytes = _get_file_raw_content(template_path)
        content_str = content_bytes.decode("utf-8", errors="ignore")
        content = Content(content_bytes, content_str, language, template_path)

        assert content is not None, f"Failed to create Content for {template_path}"

        if language_parsers := custom_parsers_by_language.get(language):
            for custom_parser in language_parsers:
                validator_result = custom_parser.validator(content)
                assert validator_result, (
                    f"Validator returned False for {template_path}. "
                    f"Expected True for helm template files."
                )

                fixed_content = custom_parser.parser(content)
                assert fixed_content is not None, (
                    f"Parser returned None for {template_path}. "
                    f"Expected a Content object after parsing."
                )



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.execution_block import (
    build_execution_block_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    if expr := match_ast_d(graph, args.n_id, "expression") or (
        expr := match_ast_d(graph, args.n_id, "comparison_expression")
    ):
        c_ids = [expr]
    else:
        c_ids = [
            n_id
            for n_id in adj_ast(graph, args.n_id)
            if graph.nodes[n_id]["label_type"] not in {"{", "}"}
        ]
    return build_execution_block_node(args, iter(c_ids))



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.json import (
    array_node as json_array,
)
from blends.syntax.readers.json import (
    boolean as json_boolean,
)
from blends.syntax.readers.json import (
    comment as json_comment,
)
from blends.syntax.readers.json import (
    document as json_document,
)
from blends.syntax.readers.json import (
    number as json_number,
)
from blends.syntax.readers.json import (
    object as json_object,
)
from blends.syntax.readers.json import (
    pair as json_pair,
)
from blends.syntax.readers.json import (
    string_node as json_string,
)

JSON_DISPATCHERS: Dispatcher = {
    "array": json_array.reader,
    "false": json_boolean.reader,
    "true": json_boolean.reader,
    "comment": json_comment.reader,
    "document": json_document.reader,
    "number": json_number.reader,
    "object": json_object.reader,
    "pair": json_pair.reader,
    "string": json_string.reader,
}



from unittest.mock import (
    patch,
)

import pytest

from blends.models import (
    Graph,
    Language,
    NId,
    SyntaxGraph,
)
from blends.syntax.builders.method_declaration import (
    build_method_declaration_node,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
    SyntaxMetadata,
)


@pytest.fixture  # type: ignore[misc,attr-defined]
def mock_args() -> SyntaxGraphArgs:  # type: ignore[misc]
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {
        "class_path": [],
        "scope_stack": ["1"],
        "next_synthetic_id": 100,
    }
    syntax_graph.add_node("1", label_type="File")

    def generic(args: SyntaxGraphArgs) -> NId:
        return args.n_id

    return SyntaxGraphArgs(
        generic=generic,
        path="test.py",
        language=Language.PYTHON,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id="2",
        metadata=metadata,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_method_declaration_emits_scope_and_definition_edges(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_method_declaration_node(
            mock_args,
            name="foo",
            direct_children={},
            list_children={},
        )

        assert mock_args.syntax_graph.has_edge("1", "100")
        def_edge_data = mock_args.syntax_graph["1"]["100"]
        assert def_edge_data["precedence"] == 1

        assert mock_args.syntax_graph.has_edge("100", "2")
        link_edge_data = mock_args.syntax_graph["100"]["2"]
        assert link_edge_data["precedence"] == 0

        assert mock_args.syntax_graph.has_edge("2", "1")
        assert not mock_args.syntax_graph.has_edge("1", "2")
        edge_data = mock_args.syntax_graph["2"]["1"]
        assert edge_data["precedence"] == 0


@pytest.mark.blends_test_group("stack_unittesting")
def test_method_declaration_does_not_emit_definition_edge_when_name_is_none(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_method_declaration_node(
            mock_args,
            name=None,
            direct_children={},
            list_children={},
        )

        assert not mock_args.syntax_graph.has_edge("1", "100")
        assert not mock_args.syntax_graph.has_edge("100", "2")
        assert mock_args.syntax_graph.has_edge("2", "1")
        assert not mock_args.syntax_graph.has_edge("1", "2")
        edge_data = mock_args.syntax_graph["2"]["1"]
        assert edge_data["precedence"] == 0


@pytest.mark.blends_test_group("stack_unittesting")
def test_method_declaration_does_not_emit_stack_graph_when_feature_preview_disabled(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", return_value=False):
        build_method_declaration_node(
            mock_args,
            name="foo",
            direct_children={},
            list_children={},
        )

        assert not mock_args.syntax_graph.has_edge("1", "100")
        assert not mock_args.syntax_graph.has_edge("100", "2")
        assert not mock_args.syntax_graph.has_edge("1", "2")
        assert not mock_args.syntax_graph.has_edge("2", "1")


@pytest.mark.blends_test_group("stack_unittesting")
def test_method_declaration_does_not_emit_edges_when_scope_stack_empty(
    mock_args: SyntaxGraphArgs,
) -> None:
    mock_args.metadata["scope_stack"] = []
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_method_declaration_node(
            mock_args,
            name="foo",
            direct_children={},
            list_children={},
        )

        assert not mock_args.syntax_graph.has_edge("1", "100")
        assert not mock_args.syntax_graph.has_edge("100", "2")
        assert "100" not in mock_args.syntax_graph.nodes
        assert not mock_args.syntax_graph.has_edge("1", "2")
        assert not mock_args.syntax_graph.has_edge("2", "1")


@pytest.mark.blends_test_group("stack_unittesting")
def test_method_declaration_preserves_scope_stack(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        initial_stack = mock_args.metadata["scope_stack"].copy()
        build_method_declaration_node(
            mock_args,
            name="foo",
            direct_children={},
            list_children={},
        )

        assert mock_args.metadata["scope_stack"] == initial_stack


@pytest.mark.blends_test_group("stack_unittesting")
def test_local_variables_in_method_connect_to_method_scope(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        build_method_declaration_node(
            mock_args,
            name="foo",
            direct_children={},
            list_children={},
        )

        mock_args.metadata["scope_stack"].append("2")

        variable_args = SyntaxGraphArgs(
            generic=mock_args.generic,
            path=mock_args.path,
            language=mock_args.language,
            ast_graph=mock_args.ast_graph,
            syntax_graph=mock_args.syntax_graph,
            n_id="3",
            metadata=mock_args.metadata,
        )
        build_variable_declaration_node(
            variable_args,
            variable_name="local_var",
            variable_type=None,
            value_id=None,
        )

        assert mock_args.syntax_graph.has_edge("2", "3")
        pop_edge_data = mock_args.syntax_graph["2"]["3"]
        assert pop_edge_data["precedence"] == 1

        assert not mock_args.syntax_graph.has_edge("1", "3")


@pytest.mark.blends_test_group("stack_unittesting")
def test_references_in_method_resolve_to_parent_scope_via_lookup(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", side_effect=lambda flag: flag == "StackGraph"):
        parent_variable_args = SyntaxGraphArgs(
            generic=mock_args.generic,
            path=mock_args.path,
            language=mock_args.language,
            ast_graph=mock_args.ast_graph,
            syntax_graph=mock_args.syntax_graph,
            n_id="3",
            metadata=mock_args.metadata,
        )
        build_variable_declaration_node(
            parent_variable_args,
            variable_name="parent_var",
            variable_type=None,
            value_id=None,
        )

        method_args = SyntaxGraphArgs(
            generic=mock_args.generic,
            path=mock_args.path,
            language=mock_args.language,
            ast_graph=mock_args.ast_graph,
            syntax_graph=mock_args.syntax_graph,
            n_id="4",
            metadata=mock_args.metadata,
        )
        build_method_declaration_node(
            method_args,
            name="bar",
            direct_children={},
            list_children={},
        )

        mock_args.metadata["scope_stack"].append("4")

        lookup_args = SyntaxGraphArgs(
            generic=mock_args.generic,
            path=mock_args.path,
            language=mock_args.language,
            ast_graph=mock_args.ast_graph,
            syntax_graph=mock_args.syntax_graph,
            n_id="5",
            metadata=mock_args.metadata,
        )
        build_symbol_lookup_node(lookup_args, symbol="parent_var")

        assert mock_args.syntax_graph.has_edge("5", "4")
        push_edge_data = mock_args.syntax_graph["5"]["4"]
        assert push_edge_data["precedence"] == 0

        assert mock_args.syntax_graph.has_edge("4", "1")
        lookup_edge_data = mock_args.syntax_graph["4"]["1"]
        assert lookup_edge_data["precedence"] == 0

        assert mock_args.syntax_graph.has_edge("1", "3")
        pop_edge_data = mock_args.syntax_graph["1"]["3"]
        assert pop_edge_data["precedence"] == 1



from collections.abc import Iterable

from blends.models import (
    NId,
    SyntaxGraph,
)
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_EXPORTED,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_PRECEDENCE,
    STACK_GRAPH_SCOPE,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
    is_stack_graph_kind,
)

_KINDS_WITH_SYMBOL = {
    StackGraphNodeKind.PUSH_SYMBOL.value,
    StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value,
    StackGraphNodeKind.POP_SYMBOL.value,
    StackGraphNodeKind.POP_SCOPED_SYMBOL.value,
}

_PUSH_KINDS = {
    StackGraphNodeKind.PUSH_SYMBOL.value,
    StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value,
}

_POP_KINDS = {
    StackGraphNodeKind.POP_SYMBOL.value,
    StackGraphNodeKind.POP_SCOPED_SYMBOL.value,
}


def _validate_kind(node_id: NId, kind: object) -> list[str]:
    if not is_stack_graph_kind(kind):
        return [f"Unknown stack graph kind {kind!r} on node {node_id!r}"]
    return []


def _validate_precedence(node_id: NId, precedence: object) -> list[str]:
    if isinstance(precedence, int):
        return []
    return [f"Invalid precedence on node {node_id!r}"]


def _validate_symbol(node_id: NId, kind: str, symbol: object) -> list[str]:
    if kind in _KINDS_WITH_SYMBOL:
        if isinstance(symbol, str) and symbol:
            return []
        return [f"Missing symbol on node {node_id!r}"]
    if symbol is None:
        return []
    return [f"Unexpected symbol on node {node_id!r}"]


def _validate_scope(node_id: NId, kind: str, scope: object) -> list[str]:
    if kind == StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value:
        if isinstance(scope, str) and scope:
            return []
        return [f"Missing scope on node {node_id!r}"]
    if scope is None:
        return []
    return [f"Unexpected scope on node {node_id!r}"]


def _validate_is_exported(node_id: NId, kind: str, is_exported: object) -> list[str]:
    if kind == StackGraphNodeKind.SCOPE.value:
        if isinstance(is_exported, bool):
            return []
        return [f"Missing is_exported on node {node_id!r}"]
    if is_exported is False or is_exported is None:
        return []
    return [f"Unexpected is_exported on node {node_id!r}"]


def _validate_is_reference(node_id: NId, kind: str, is_reference: object) -> list[str]:
    if kind in _PUSH_KINDS:
        if isinstance(is_reference, bool):
            return []
        return [f"Missing is_reference on node {node_id!r}"]
    if is_reference is False or is_reference is None:
        return []
    return [f"Unexpected is_reference on node {node_id!r}"]


def _validate_is_definition(node_id: NId, kind: str, is_definition: object) -> list[str]:
    if kind in _POP_KINDS:
        if isinstance(is_definition, bool):
            return []
        return [f"Missing is_definition on node {node_id!r}"]
    if is_definition is False or is_definition is None:
        return []
    return [f"Unexpected is_definition on node {node_id!r}"]


def validate_stack_graph_node(graph: SyntaxGraph, node_id: NId) -> list[str]:
    node_data = graph.nodes.get(node_id, {})
    kind = node_data.get(STACK_GRAPH_KIND)
    if kind is None:
        return []

    kind_errors = _validate_kind(node_id, kind)
    if kind_errors:
        return kind_errors

    precedence = node_data.get(STACK_GRAPH_PRECEDENCE, 0)
    symbol = node_data.get(STACK_GRAPH_SYMBOL)
    scope = node_data.get(STACK_GRAPH_SCOPE)
    is_exported = node_data.get(STACK_GRAPH_IS_EXPORTED)
    is_reference = node_data.get(STACK_GRAPH_IS_REFERENCE)
    is_definition = node_data.get(STACK_GRAPH_IS_DEFINITION)

    return [
        *_validate_precedence(node_id, precedence),
        *_validate_symbol(node_id, kind, symbol),
        *_validate_scope(node_id, kind, scope),
        *_validate_is_exported(node_id, kind, is_exported),
        *_validate_is_reference(node_id, kind, is_reference),
        *_validate_is_definition(node_id, kind, is_definition),
    ]


def validate_stack_graph_graph(
    graph: SyntaxGraph, *, node_ids: Iterable[NId] | None = None
) -> list[str]:
    ids = list(node_ids) if node_ids is not None else list(graph.nodes)
    return [err for node_id in ids for err in validate_stack_graph_node(graph, node_id)]



import logging

from blends.models import (
    NId,
    SyntaxGraph,
)
from blends.syntax.cfg.dispatchers import (
    DISPATCHERS,
)
from blends.syntax.models import (
    MissingCfgBuilderError,
    SyntaxCfgArgs,
)

LOGGER = logging.getLogger(__name__)


def generic(args: SyntaxCfgArgs) -> NId:
    node_type = args.graph.nodes[args.n_id]["label_type"]

    if dispatcher := DISPATCHERS.get(node_type):
        return dispatcher(args)

    exc_log = f"Missing cfg builder for {node_type}"
    raise MissingCfgBuilderError(exc_log)


def add_syntax_cfg(graph: SyntaxGraph, *, is_multifile: bool = False) -> SyntaxGraph:
    try:
        generic(
            args=SyntaxCfgArgs(
                generic,
                graph,
                n_id=NId(1),
                nxt_id=None,
                is_multifile=is_multifile,
            ),
        )
    except MissingCfgBuilderError:
        LOGGER.exception("Missing cfg builder")
    return graph



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_d,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_id = match_ast_d(graph, args.n_id, "variable_declaration")
    mod_n = match_ast_d(graph, args.n_id, "modifier")
    c_mod_n = None
    if (mod_n := match_ast_d(graph, args.n_id, "modifier")) and (adj := adj_ast(graph, mod_n)):
        c_mod_n = node_to_str(graph, adj[0])

    if not var_id:
        var_id = args.n_id

    var_type_id = graph.nodes[var_id]["label_field_type"]
    var_type = node_to_str(graph, var_type_id)

    var_decl_id = match_ast_d(graph, var_id, "variable_declarator")
    if not var_decl_id:
        var_decl_id = adj_ast(graph, var_id)[0]

    childs = match_ast(graph, var_decl_id, "identifier")
    if identifier_id := childs.get("identifier"):
        var_name = node_to_str(graph, identifier_id)
    else:
        var_name = node_to_str(graph, var_decl_id)

    value_id = None
    if (declarator_id := match_ast_d(graph, var_id, "variable_declarator")) and (
        declarator_childs := list(
            filter(
                lambda child: graph.nodes[child]["label_type"] not in ("identifier", "="),
                list(adj_ast(graph, declarator_id)),
            ),
        )
    ):
        value_id = declarator_childs[0]

    return build_variable_declaration_node(args, var_name, var_type, value_id, None, c_mod_n)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.import_global import (
    build_import_global_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes

    name_n_id = match_ast_d(graph, args.n_id, "identifier") or match_ast_d(
        graph,
        args.n_id,
        "qualified_name",
    )
    expression = node_to_str(graph, name_n_id) if name_n_id else node_to_str(graph, args.n_id)

    alias: str | None = None
    if (alias_n_id := nodes[args.n_id].get("label_field_alias")) and (
        ident_n_id := match_ast_d(graph, alias_n_id, "identifier")
    ):
        alias = nodes[ident_n_id].get("label_text")

    return build_import_global_node(args, expression, set(), alias)



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_ternary_operation_node(
    args: SyntaxGraphArgs,
    condition_id: NId,
    alternative_id: NId,
    consequence_id: NId,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        condition_id=condition_id,
        true_id=alternative_id,
        false_id=consequence_id,
        label_type="TernaryOperation",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(condition_id)),
        label_ast="AST",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(alternative_id)),
        label_ast="AST",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(consequence_id)),
        label_ast="AST",
    )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.missing_node import (
    build_missing_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs, n_type: str) -> NId:
    c_ids = adj_ast(args.ast_graph, args.n_id)

    return build_missing_node(
        args,
        n_type,
        iter(c_ids),
    )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.stack.stacks import (
    ScopeStackNode,
    StackState,
    SymbolStackNode,
)

if TYPE_CHECKING:
    from blends.stack.view import (
        StackGraphView,
    )


class TransitionError(str, Enum):
    EMPTY_SCOPE_STACK = "EmptyScopeStack"
    EMPTY_SYMBOL_STACK = "EmptySymbolStack"
    INCORRECT_POPPED_SYMBOL = "IncorrectPoppedSymbol"
    MISSING_ATTACHED_SCOPE_LIST = "MissingAttachedScopeList"
    UNEXPECTED_ATTACHED_SCOPE_LIST = "UnexpectedAttachedScopeList"
    UNKNOWN_ATTACHED_SCOPE = "UnknownAttachedScope"
    MISSING_SCOPE_PAYLOAD = "MissingScopePayload"
    MISSING_SYMBOL_PAYLOAD = "MissingSymbolPayload"
    UNKNOWN_NODE_KIND = "UnknownNodeKind"


@dataclass(frozen=True, slots=True)
class TransitionResult:
    state: StackState
    jump_target: int | None
    error: TransitionError | None


def _handle_push_symbol(
    view: StackGraphView, node_index: int, state: StackState
) -> TransitionResult:
    sym = view.symbol_id_at(node_index)
    if sym is None:
        return TransitionResult(state, None, TransitionError.MISSING_SYMBOL_PAYLOAD)
    new_symbol_stack = SymbolStackNode(symbol_id=sym, scopes=None, tail=state.symbol_stack)
    return TransitionResult(StackState(new_symbol_stack, state.scope_stack), None, None)


def _handle_pop_symbol(
    view: StackGraphView, node_index: int, state: StackState
) -> TransitionResult:
    sym = view.symbol_id_at(node_index)
    if sym is None:
        return TransitionResult(state, None, TransitionError.MISSING_SYMBOL_PAYLOAD)
    top = state.symbol_stack
    if top is None:
        return TransitionResult(state, None, TransitionError.EMPTY_SYMBOL_STACK)
    if top.symbol_id != sym:
        return TransitionResult(state, None, TransitionError.INCORRECT_POPPED_SYMBOL)
    if top.scopes is not None:
        return TransitionResult(state, None, TransitionError.UNEXPECTED_ATTACHED_SCOPE_LIST)
    return TransitionResult(StackState(top.tail, state.scope_stack), None, None)


def _handle_drop_scopes(
    _view: StackGraphView, _node_index: int, state: StackState
) -> TransitionResult:
    return TransitionResult(StackState(state.symbol_stack, None), None, None)


def _handle_jump_to(_view: StackGraphView, _node_index: int, state: StackState) -> TransitionResult:
    top = state.scope_stack
    if top is None:
        return TransitionResult(state, None, TransitionError.EMPTY_SCOPE_STACK)
    jump_target = top.scope_index
    return TransitionResult(StackState(state.symbol_stack, top.tail), jump_target, None)


def _handle_push_scoped_symbol(
    view: StackGraphView, node_index: int, state: StackState
) -> TransitionResult:
    sym = view.symbol_id_at(node_index)
    if sym is None:
        return TransitionResult(state, None, TransitionError.MISSING_SYMBOL_PAYLOAD)
    attached_scope = view.scope_index_at(node_index)
    if attached_scope is None:
        return TransitionResult(state, None, TransitionError.MISSING_SCOPE_PAYLOAD)
    if attached_scope not in view.exported_scopes:
        return TransitionResult(state, None, TransitionError.UNKNOWN_ATTACHED_SCOPE)
    attached_scopes = ScopeStackNode(scope_index=attached_scope, tail=state.scope_stack)
    new_symbol_stack = SymbolStackNode(
        symbol_id=sym,
        scopes=attached_scopes,
        tail=state.symbol_stack,
    )
    return TransitionResult(StackState(new_symbol_stack, state.scope_stack), None, None)


def _handle_pop_scoped_symbol(
    view: StackGraphView, node_index: int, state: StackState
) -> TransitionResult:
    sym = view.symbol_id_at(node_index)
    if sym is None:
        return TransitionResult(state, None, TransitionError.MISSING_SYMBOL_PAYLOAD)
    top = state.symbol_stack
    if top is None:
        return TransitionResult(state, None, TransitionError.EMPTY_SYMBOL_STACK)
    if top.symbol_id != sym:
        return TransitionResult(state, None, TransitionError.INCORRECT_POPPED_SYMBOL)
    if top.scopes is None:
        return TransitionResult(state, None, TransitionError.MISSING_ATTACHED_SCOPE_LIST)
    return TransitionResult(StackState(top.tail, top.scopes), None, None)


def _handle_noop(_view: StackGraphView, _node_index: int, state: StackState) -> TransitionResult:
    return TransitionResult(state, None, None)


def apply_node(view: StackGraphView, node_index: int, state: StackState) -> TransitionResult:
    kind = view.kind_at(node_index)
    handler = {
        StackGraphNodeKind.DROP_SCOPES.value: _handle_drop_scopes,
        StackGraphNodeKind.JUMP_TO.value: _handle_jump_to,
        StackGraphNodeKind.POP_SCOPED_SYMBOL.value: _handle_pop_scoped_symbol,
        StackGraphNodeKind.POP_SYMBOL.value: _handle_pop_symbol,
        StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value: _handle_push_scoped_symbol,
        StackGraphNodeKind.PUSH_SYMBOL.value: _handle_push_symbol,
        StackGraphNodeKind.ROOT.value: _handle_noop,
        StackGraphNodeKind.SCOPE.value: _handle_noop,
    }.get(kind)
    if handler is None:
        return TransitionResult(state, None, TransitionError.UNKNOWN_NODE_KIND)
    return handler(view, node_index, state)



from blends.models import (
    NId,
)
from blends.syntax.builders.namespace import (
    build_namespace_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    if label_n_id := graph.nodes[args.n_id].get("label_field_name"):
        name = node_to_str(graph, label_n_id).replace("\\", "/")
    else:
        name = "anonymous_namespace"

    block_id = graph.nodes[args.n_id].get("label_field_body")

    return build_namespace_node(args, name, block_id)



from collections.abc import (
    Iterator,
)

from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_selector_node(
    args: SyntaxGraphArgs,
    identifier: str | None,
    c_ids: Iterator[NId] | None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="Selector",
    )

    if identifier:
        args.syntax_graph.update_node_attribute(args.n_id, "selector_name", identifier)

    if c_ids:
        for c_id in c_ids:
            args.syntax_graph.add_edge(
                args.n_id,
                args.generic(args.fork_n_id(c_id)),
                label_ast="AST",
            )

    return args.n_id



import pytest

from blends.models import Language, SyntaxGraph
from blends.stack.policies.base import ExportPolicy
from blends.stack.policies.default import DefaultExportPolicy
from blends.stack.policies.go import GoExportPolicy
from blends.stack.policies.java import JavaExportPolicy
from blends.stack.policies.javascript import JavaScriptExportPolicy
from blends.stack.policies.kotlin import KotlinExportPolicy
from blends.stack.policies.python import PythonExportPolicy
from blends.stack.policies.registry import get_export_policy
from blends.stack.scope_types import ScopeType


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    ("language", "expected_type"),
    [
        (Language.PYTHON, PythonExportPolicy),
        (Language.JAVA, JavaExportPolicy),
        (Language.GO, GoExportPolicy),
        (Language.JAVASCRIPT, JavaScriptExportPolicy),
        (Language.TYPESCRIPT, JavaScriptExportPolicy),
        (Language.KOTLIN, KotlinExportPolicy),
        (Language.SCALA, KotlinExportPolicy),
        (Language.RUBY, DefaultExportPolicy),
    ],
)
def test_registry_returns_correct_policy(  # type: ignore[misc]
    language: Language, expected_type: type[ExportPolicy]
) -> None:
    assert isinstance(get_export_policy(language), expected_type)


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    "policy",
    [
        PythonExportPolicy(),
        JavaExportPolicy(),
        GoExportPolicy(),
        JavaScriptExportPolicy(),
        DefaultExportPolicy(),
    ],
)
def test_file_scope_always_exported(policy: ExportPolicy) -> None:  # type: ignore[misc]
    graph = SyntaxGraph()
    graph.add_node("1")
    assert policy.is_scope_exported(graph, "1", ScopeType.FILE.value) is True


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    ("policy", "scope_type"),
    [
        (PythonExportPolicy(), ScopeType.FUNCTION.value),
        (PythonExportPolicy(), ScopeType.CLASS.value),
        (JavaScriptExportPolicy(), ScopeType.FUNCTION.value),
        (DefaultExportPolicy(), ScopeType.CLASS.value),
    ],
)
def test_non_file_scope_not_exported(  # type: ignore[misc]
    policy: ExportPolicy, scope_type: str
) -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    assert policy.is_scope_exported(graph, "1", scope_type) is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_java_policy_public_class_scope_exported() -> None:
    policy = JavaExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1", access_modifiers="public")
    assert policy.is_scope_exported(graph, "1", ScopeType.CLASS.value) is True


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    "modifiers", ["private", None]
)
def test_java_policy_non_public_class_scope_not_exported(  # type: ignore[misc]
    modifiers: str | None,
) -> None:
    policy = JavaExportPolicy()
    graph = SyntaxGraph()
    if modifiers:
        graph.add_node("1", access_modifiers=modifiers)
    else:
        graph.add_node("1")
    assert policy.is_scope_exported(graph, "1", ScopeType.CLASS.value) is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_kotlin_policy_class_without_modifier_exported() -> None:
    policy = KotlinExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1")
    assert policy.is_scope_exported(graph, "1", ScopeType.CLASS.value) is True


@pytest.mark.blends_test_group("stack_unittesting")
def test_kotlin_policy_private_class_not_exported() -> None:
    policy = KotlinExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1", access_modifiers="private")
    assert policy.is_scope_exported(graph, "1", ScopeType.CLASS.value) is False


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    "policy",
    [PythonExportPolicy(), JavaScriptExportPolicy(), DefaultExportPolicy()],
)
def test_scope_based_top_level_definition_exported(  # type: ignore[misc]
    policy: ExportPolicy,
) -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    assert policy.is_definition_exported(graph, "1", None) is True


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    "policy",
    [PythonExportPolicy(), JavaScriptExportPolicy(), DefaultExportPolicy()],
)
def test_scope_based_file_level_definition_exported(  # type: ignore[misc]
    policy: ExportPolicy,
) -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.add_node("2", stack_graph_scope_type=ScopeType.FILE.value)
    assert policy.is_definition_exported(graph, "1", "2") is True


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    "policy",
    [PythonExportPolicy(), JavaScriptExportPolicy(), DefaultExportPolicy()],
)
def test_scope_based_function_level_definition_not_exported(  # type: ignore[misc]
    policy: ExportPolicy,
) -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.add_node("2", stack_graph_scope_type=ScopeType.FUNCTION.value)
    assert policy.is_definition_exported(graph, "1", "2") is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_java_policy_public_definition_in_exported_scope() -> None:
    policy = JavaExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1", access_modifiers="public")
    graph.add_node("2", stack_graph_is_exported=True)
    assert policy.is_definition_exported(graph, "1", "2") is True


@pytest.mark.blends_test_group("stack_unittesting")
def test_java_policy_private_definition_not_exported() -> None:
    policy = JavaExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1", access_modifiers="private")
    graph.add_node("2", stack_graph_is_exported=True)
    assert policy.is_definition_exported(graph, "1", "2") is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_java_policy_public_definition_in_non_exported_scope() -> None:
    policy = JavaExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1", access_modifiers="public")
    graph.add_node("2", stack_graph_is_exported=False)
    assert policy.is_definition_exported(graph, "1", "2") is False


@pytest.mark.blends_test_group("stack_unittesting")
def test_kotlin_policy_definition_without_modifier_exported() -> None:
    policy = KotlinExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1")
    graph.add_node("2", stack_graph_is_exported=True)
    assert policy.is_definition_exported(graph, "1", "2") is True


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    ("node_attrs", "expected"),
    [
        ({"stack_graph_symbol": "ExportedFunc"}, True),
        ({"stack_graph_symbol": "privateFunc"}, False),
        ({"name": "ExportedVar"}, True),
        ({"name": "privateVar"}, False),
    ],
)
def test_go_policy_top_level_capitalization(  # type: ignore[misc]
    node_attrs: dict[str, str],
    expected: bool,  # noqa: FBT001
) -> None:
    policy = GoExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1", **node_attrs)
    assert policy.is_definition_exported(graph, "1", None) is expected


@pytest.mark.blends_test_group("stack_unittesting")
@pytest.mark.parametrize(  # type: ignore[misc,attr-defined]
    ("scope_type", "expected"),
    [
        (ScopeType.FUNCTION.value, False),
        (ScopeType.FILE.value, True),
    ],
)
def test_go_policy_capitalized_in_scope(  # type: ignore[misc]
    scope_type: str,
    expected: bool,  # noqa: FBT001
) -> None:
    policy = GoExportPolicy()
    graph = SyntaxGraph()
    graph.add_node("1", stack_graph_symbol="ExportedFunc")
    graph.add_node("2", stack_graph_scope_type=scope_type)
    assert policy.is_definition_exported(graph, "1", "2") is expected



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.array_node import (
    build_array_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    childs_id = adj_ast(
        args.ast_graph,
        args.n_id,
    )

    node_types = {
        "array_initializer",
        "binary_expression",
        "cast_expression",
        "decimal_integer_literal",
        "identifier",
        "field_access",
        "object",
        "object_creation_expression",
        "string",
        "string_literal",
    }

    valid_childs = [
        child for child in childs_id if args.ast_graph.nodes[child]["label_type"] in node_types
    ]

    return build_array_node(args, valid_childs)



from blends.models import (
    NId,
)
from blends.query import (
    get_nodes_by_path,
    match_ast_group_d,
)
from blends.syntax.builders.import_statement import (
    build_import_statement_node,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def process_ruby_imports(args: SyntaxGraphArgs, arguments_id: NId, method_name: str) -> NId:
    graph = args.ast_graph
    module_nodes: set[NId] = set()
    first_path = get_nodes_by_path(graph, arguments_id, "string", "string_content")
    second_path = get_nodes_by_path(graph, arguments_id, "string", "interpolation", "identifier")
    arg = match_ast_group_d(graph, arguments_id, "identifier")
    module_nodes.update(first_path)
    module_nodes.update(second_path)
    module_nodes.update(arg)
    import_attrs: list[dict[str, str]] = []
    for module_id in module_nodes:
        module_expression = node_to_str(graph, module_id)
        import_attrs.append(
            {
                "corrected_n_id": module_id,
                "method_name": method_name,
                "expression": module_expression,
            },
        )
    return build_import_statement_node(args, *import_attrs)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    node = graph.nodes[args.n_id]

    arguments_id = node.get("label_field_arguments")
    method_id = node.get("label_field_method")
    operator_id = node.get("label_field_operator")
    object_id = node.get("label_field_receiver")
    block_id = node.get("label_field_block")

    method_name = node_to_str(graph, method_id) if method_id else ""

    if operator_id and node_to_str(graph, operator_id) == "::":
        method_name = f"::{method_name}"

    ruby_imports = {"require", "require_relative", "load", "autoload"}
    if method_name in ruby_imports and arguments_id and not object_id:
        return process_ruby_imports(args, arguments_id, method_name)

    children: DirectChildren = {}
    if method_id:
        children["expression_id"] = method_id
    if arguments_id:
        children["arguments_id"] = arguments_id
    if object_id:
        children["object_id"] = object_id
    if block_id:
        children["block_id"] = block_id

    return build_method_invocation_node(args, method_name, children)



from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    args.evaluation[args.n_id] = False
    block_id = args.graph.nodes[args.n_id]["block_id"]
    args.evaluation[args.n_id] = generic(args.fork_n_id(block_id)).danger

    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.try_statement import (
    build_try_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    try_node = graph.nodes[args.n_id]
    block_id = try_node["label_field_body"]

    catch_clauses = match_ast_group_d(graph, args.n_id, "block")
    if len(catch_clauses) > 0 and catch_clauses[0] == block_id:
        catch_clauses.pop(0)

    if finally_clause := match_ast_d(graph, args.n_id, "finally_clause"):
        catch_clauses.append(finally_clause)

    return build_try_statement_node(args, block_id, catch_clauses, None, None)



from collections import deque
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import NoReturn

from blends.stack.criteria import (
    is_endpoint,
    is_jump_to_boundary,
)
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path.errors import (
    PartialPathResolutionError,
    _raise_incorrect_popped_symbol,
    _raise_missing_attached_scope_list,
    _raise_symbol_stack_unsatisfied,
    _raise_unexpected_attached_scope_list,
    _raise_unknown_attached_scope,
)
from blends.stack.partial_path.partial_path import PartialPath
from blends.stack.partial_path.partial_stacks import (
    PartialScopedSymbol,
    PartialScopeStack,
    PartialSymbolStack,
)
from blends.stack.partial_path.stack_conditions import (
    StackConditions,
    create_seed_path,
)
from blends.stack.partial_path.variables import ScopeStackVariable
from blends.stack.selection import PartialPathEdge
from blends.stack.stacks import (
    ScopeStackNode,
    StackState,
    SymbolStackNode,
)
from blends.stack.transitions import apply_node
from blends.stack.view import StackGraphView


@dataclass(frozen=True, slots=True)
class PartialPathLimits:
    max_work_per_phase: int | None = None
    enable_dedupe: bool = True
    enable_cycle_guard: bool = True


@dataclass(slots=True)
class PerFilePartialPathStats:
    seed_count: int = 0
    explored_count: int = 0
    accepted_count: int = 0
    deduped_count: int = 0
    discarded_count: int = 0
    max_queue_size: int = 0
    limit_hits: dict[str, int] = field(default_factory=dict)
    length_histogram: dict[int, int] = field(default_factory=dict)

    def record_length(self, length: int) -> None:
        self.length_histogram[length] = self.length_histogram.get(length, 0) + 1

    def mark_limit_hit(self, name: str) -> None:
        self.limit_hits[name] = self.limit_hits.get(name, 0) + 1


@dataclass(slots=True)
class PerFilePartialPaths:
    paths: tuple[PartialPath, ...]
    stats: PerFilePartialPathStats


def _collect_seed_nodes(view: StackGraphView, file_path: str) -> list[int]:
    file_nodes = view.file_to_nodes.get(file_path, frozenset())
    seeds = [
        node_index
        for node_index in file_nodes
        if is_endpoint(view, node_index) or is_jump_to_boundary(view, node_index)
    ]
    if 0 not in seeds:
        seeds.append(0)
    return sorted(seeds)


def _build_scope_stack(scopes: tuple[int, ...]) -> ScopeStackNode | None:
    tail: ScopeStackNode | None = None
    for scope_index in reversed(scopes):
        tail = ScopeStackNode(scope_index=scope_index, tail=tail)
    return tail


def _build_symbol_stack(
    symbols: tuple[PartialScopedSymbol, ...],
) -> tuple[SymbolStackNode | None, bool]:
    tail: SymbolStackNode | None = None
    for symbol in reversed(symbols):
        scopes = symbol.scopes
        if scopes is None:
            scopes_node = None
        else:
            if scopes.variable is not None:
                return None, False
            scopes_node = _build_scope_stack(scopes.scopes)
        tail = SymbolStackNode(symbol_id=symbol.symbol_id, scopes=scopes_node, tail=tail)
    return tail, True


def _build_concrete_state(
    symbol_stack: PartialSymbolStack, scope_stack: PartialScopeStack
) -> StackState | None:
    if symbol_stack.variable is not None or scope_stack.variable is not None:
        return None
    symbol_node, symbol_is_concrete = _build_symbol_stack(symbol_stack.symbols)
    if not symbol_is_concrete:
        return None
    scope_node = _build_scope_stack(scope_stack.scopes)
    return StackState(symbol_stack=symbol_node, scope_stack=scope_node)


def _can_apply_transition(
    view: StackGraphView,
    node_index: int,
    symbol_stack: PartialSymbolStack,
    scope_stack: PartialScopeStack,
) -> bool:
    state = _build_concrete_state(symbol_stack, scope_stack)
    if state is None:
        return True
    result = apply_node(view, node_index, state)
    return result.error is None


def _pop_symbol_stack(
    symbol_stack: PartialSymbolStack,
) -> tuple[PartialScopedSymbol | None, PartialSymbolStack]:
    if not symbol_stack.symbols:
        return None, symbol_stack
    head = symbol_stack.symbols[0]
    tail = PartialSymbolStack(symbols=symbol_stack.symbols[1:], variable=symbol_stack.variable)
    return head, tail


def _push_symbol_front(
    symbol_stack: PartialSymbolStack,
    symbol: PartialScopedSymbol,
) -> PartialSymbolStack:
    return PartialSymbolStack(
        symbols=(symbol, *symbol_stack.symbols), variable=symbol_stack.variable
    )


def _push_symbol_back(
    symbol_stack: PartialSymbolStack,
    symbol: PartialScopedSymbol,
) -> PartialSymbolStack:
    return PartialSymbolStack(
        symbols=(*symbol_stack.symbols, symbol), variable=symbol_stack.variable
    )


def _push_scope_front(scope_stack: PartialScopeStack, scope_index: int) -> PartialScopeStack:
    return PartialScopeStack(
        scopes=(scope_index, *scope_stack.scopes), variable=scope_stack.variable
    )


def _fresh_scope_stack_variable(stacks: StackConditions) -> ScopeStackVariable:
    max_var = max(
        stacks.symbol_pre.largest_scope_stack_var(),
        stacks.symbol_post.largest_scope_stack_var(),
        stacks.scope_pre.largest_scope_stack_var(),
        stacks.scope_post.largest_scope_stack_var(),
    )
    return ScopeStackVariable(var_id=max_var + 1)


def _symbol_stack_unsatisfied() -> NoReturn:
    _raise_symbol_stack_unsatisfied()


def _append_drop_scopes(
    _view: StackGraphView, _node_index: int, stacks: StackConditions
) -> StackConditions:
    return StackConditions(
        symbol_pre=stacks.symbol_pre,
        symbol_post=stacks.symbol_post,
        scope_pre=stacks.scope_pre,
        scope_post=PartialScopeStack.empty(),
    )


def _append_pop_scoped_symbol(
    view: StackGraphView, node_index: int, stacks: StackConditions
) -> StackConditions:
    symbol_id = view.symbol_id_at(node_index)
    if symbol_id is None:
        _raise_incorrect_popped_symbol()
    popped, symbol_post = _pop_symbol_stack(stacks.symbol_post)
    if popped is not None:
        if popped.symbol_id != symbol_id:
            _raise_incorrect_popped_symbol()
        if popped.scopes is None:
            _raise_missing_attached_scope_list()
        return StackConditions(
            symbol_pre=stacks.symbol_pre,
            symbol_post=symbol_post,
            scope_pre=stacks.scope_pre,
            scope_post=popped.scopes,
        )
    if stacks.symbol_post.has_variable():
        scope_var = _fresh_scope_stack_variable(stacks)
        attached_scopes = PartialScopeStack.from_variable(scope_var)
        symbol_pre = _push_symbol_back(
            stacks.symbol_pre,
            PartialScopedSymbol(symbol_id=symbol_id, scopes=attached_scopes),
        )
        return StackConditions(
            symbol_pre=symbol_pre,
            symbol_post=symbol_post,
            scope_pre=stacks.scope_pre,
            scope_post=PartialScopeStack.from_variable(scope_var),
        )
    return _symbol_stack_unsatisfied()


def _append_pop_symbol(
    view: StackGraphView, node_index: int, stacks: StackConditions
) -> StackConditions:
    symbol_id = view.symbol_id_at(node_index)
    if symbol_id is None:
        _raise_incorrect_popped_symbol()
    popped, symbol_post = _pop_symbol_stack(stacks.symbol_post)
    if popped is not None:
        if popped.symbol_id != symbol_id:
            _raise_incorrect_popped_symbol()
        if popped.scopes is not None:
            _raise_unexpected_attached_scope_list()
        return StackConditions(
            symbol_pre=stacks.symbol_pre,
            symbol_post=symbol_post,
            scope_pre=stacks.scope_pre,
            scope_post=stacks.scope_post,
        )
    if stacks.symbol_post.has_variable():
        symbol_pre = _push_symbol_back(
            stacks.symbol_pre, PartialScopedSymbol(symbol_id=symbol_id, scopes=None)
        )
        return StackConditions(
            symbol_pre=symbol_pre,
            symbol_post=symbol_post,
            scope_pre=stacks.scope_pre,
            scope_post=stacks.scope_post,
        )
    return _symbol_stack_unsatisfied()


def _append_push_scoped_symbol(
    view: StackGraphView, node_index: int, stacks: StackConditions
) -> StackConditions:
    symbol_id = view.symbol_id_at(node_index)
    scope_index = view.scope_index_at(node_index)
    if symbol_id is None:
        _raise_incorrect_popped_symbol()
    if scope_index is None:
        _raise_unknown_attached_scope()
    attached_scopes = _push_scope_front(stacks.scope_post, scope_index)
    symbol_post = _push_symbol_front(
        stacks.symbol_post,
        PartialScopedSymbol(symbol_id=symbol_id, scopes=attached_scopes),
    )
    return StackConditions(
        symbol_pre=stacks.symbol_pre,
        symbol_post=symbol_post,
        scope_pre=stacks.scope_pre,
        scope_post=stacks.scope_post,
    )


def _append_push_symbol(
    view: StackGraphView, node_index: int, stacks: StackConditions
) -> StackConditions:
    symbol_id = view.symbol_id_at(node_index)
    if symbol_id is None:
        _raise_incorrect_popped_symbol()
    symbol_post = _push_symbol_front(
        stacks.symbol_post,
        PartialScopedSymbol(symbol_id=symbol_id, scopes=None),
    )
    return StackConditions(
        symbol_pre=stacks.symbol_pre,
        symbol_post=symbol_post,
        scope_pre=stacks.scope_pre,
        scope_post=stacks.scope_post,
    )


_APPEND_BY_KIND: dict[str, Callable[[StackGraphView, int, StackConditions], StackConditions]] = {
    StackGraphNodeKind.DROP_SCOPES.value: _append_drop_scopes,
    StackGraphNodeKind.POP_SCOPED_SYMBOL.value: _append_pop_scoped_symbol,
    StackGraphNodeKind.POP_SYMBOL.value: _append_pop_symbol,
    StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value: _append_push_scoped_symbol,
    StackGraphNodeKind.PUSH_SYMBOL.value: _append_push_symbol,
}


def _append_node_to_partial_stacks(
    view: StackGraphView, node_index: int, stacks: StackConditions
) -> StackConditions:
    kind = view.kind_at(node_index)
    handler = _APPEND_BY_KIND.get(kind)
    if handler is None:
        return stacks
    return handler(view, node_index, stacks)


def _extend_path_with_edge(
    view: StackGraphView,
    current_path: PartialPath,
    sink_node_index: int,
    precedence: int,
    *,
    record_edges: bool,
) -> PartialPath:
    stacks = _append_node_to_partial_stacks(
        view,
        sink_node_index,
        StackConditions(
            symbol_pre=current_path.symbol_stack_precondition,
            symbol_post=current_path.symbol_stack_postcondition,
            scope_pre=current_path.scope_stack_precondition,
            scope_post=current_path.scope_stack_postcondition,
        ),
    )
    if record_edges:
        edges = (
            *current_path.edges,
            PartialPathEdge(source_node_index=current_path.end_node_index, precedence=precedence),
        )
    else:
        edges = current_path.edges
    return PartialPath(
        start_node_index=current_path.start_node_index,
        end_node_index=sink_node_index,
        symbol_stack_precondition=stacks.symbol_pre,
        symbol_stack_postcondition=stacks.symbol_post,
        scope_stack_precondition=stacks.scope_pre,
        scope_stack_postcondition=stacks.scope_post,
        edges=edges,
    )


def _should_accept_path(view: StackGraphView, path: PartialPath, file_path: str) -> bool:
    if not (
        is_endpoint(view, path.start_node_index) or is_jump_to_boundary(view, path.start_node_index)
    ):
        return False
    file_nodes = view.file_to_nodes.get(file_path, frozenset())
    return (
        path.end_node_index in file_nodes and is_endpoint(view, path.end_node_index)
    ) or is_jump_to_boundary(view, path.end_node_index)


@dataclass(slots=True)
class _ExpansionContext:
    view: StackGraphView
    file_path: str
    limits: PartialPathLimits
    stats: PerFilePartialPathStats
    accepted_paths: list[PartialPath]
    seen_paths: set[
        tuple[
            int,
            int,
            PartialSymbolStack,
            PartialSymbolStack,
            PartialScopeStack,
            PartialScopeStack,
        ]
    ]
    seen_states: set[tuple[int, int, PartialSymbolStack, PartialScopeStack]]
    queue: deque[tuple[PartialPath, int]]
    record_edges: bool


def _process_current_path(
    context: _ExpansionContext, current_path: PartialPath, current_length: int
) -> int:
    if current_length > 0 and _should_accept_path(context.view, current_path, context.file_path):
        path_key = (
            current_path.start_node_index,
            current_path.end_node_index,
            current_path.symbol_stack_precondition,
            current_path.symbol_stack_postcondition,
            current_path.scope_stack_precondition,
            current_path.scope_stack_postcondition,
        )
        if context.limits.enable_dedupe and path_key in context.seen_paths:
            context.stats.deduped_count += 1
            return 0
        context.seen_paths.add(path_key)
        context.accepted_paths.append(current_path)
        context.stats.accepted_count += 1
        context.stats.record_length(current_length)
        return 0

    if context.limits.enable_cycle_guard:
        state_key = (
            current_path.start_node_index,
            current_path.end_node_index,
            current_path.symbol_stack_postcondition,
            current_path.scope_stack_postcondition,
        )
        if state_key in context.seen_states:
            context.stats.discarded_count += 1
            return 0
        context.seen_states.add(state_key)

    outgoing_edges = context.view.outgoing[current_path.end_node_index]
    for sink_node_index, precedence in outgoing_edges:
        if not _can_apply_transition(
            context.view,
            sink_node_index,
            current_path.symbol_stack_postcondition,
            current_path.scope_stack_postcondition,
        ):
            context.stats.discarded_count += 1
            continue
        try:
            extended_path = _extend_path_with_edge(
                context.view,
                current_path,
                sink_node_index,
                precedence,
                record_edges=context.record_edges,
            )
            context.queue.append((extended_path, current_length + 1))
        except PartialPathResolutionError:
            context.stats.discarded_count += 1
            continue

    return len(outgoing_edges)


def compute_minimal_partial_paths_in_file(
    view: StackGraphView,
    file_path: str,
    *,
    limits: PartialPathLimits | None = None,
    record_edges: bool = True,
) -> PerFilePartialPaths:
    if limits is None:
        limits = PartialPathLimits()
    stats = PerFilePartialPathStats()
    accepted_paths: list[PartialPath] = []
    seen_paths: set[
        tuple[
            int,
            int,
            PartialSymbolStack,
            PartialSymbolStack,
            PartialScopeStack,
            PartialScopeStack,
        ]
    ] = set()
    seen_states: set[tuple[int, int, PartialSymbolStack, PartialScopeStack]] = set()

    seed_nodes = _collect_seed_nodes(view, file_path)
    stats.seed_count = len(seed_nodes)

    queue: deque[tuple[PartialPath, int]] = deque()
    for seed_node_index in seed_nodes:
        seed_path = create_seed_path(view, seed_node_index)
        queue.append((seed_path, 0))

    context = _ExpansionContext(
        view=view,
        file_path=file_path,
        limits=limits,
        stats=stats,
        accepted_paths=accepted_paths,
        seen_paths=seen_paths,
        seen_states=seen_states,
        queue=queue,
        record_edges=record_edges,
    )

    work_performed = 0
    while queue:
        stats.max_queue_size = max(stats.max_queue_size, len(queue))

        if limits.max_work_per_phase is not None and work_performed >= limits.max_work_per_phase:
            stats.mark_limit_hit("max_work_per_phase")
            break

        current_path, current_length = queue.popleft()
        stats.explored_count += 1
        candidate_count = _process_current_path(context, current_path, current_length)
        work_performed += candidate_count

    return PerFilePartialPaths(paths=tuple(accepted_paths), stats=stats)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph, n_id = args.ast_graph, args.n_id
    text = node_to_str(graph, n_id)

    interpolations = match_ast_group_d(graph, n_id, "interpolation")
    valid_nodes = [
        child_id
        for interp in interpolations
        for child_id in adj_ast(graph, interp)
        if graph.nodes[child_id]["label_type"] in {"identifier", "element_reference"}
    ]

    return (
        build_string_literal_node(args, text, iter(valid_nodes))
        if valid_nodes
        else build_string_literal_node(args, text)
    )
